# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains DataFrameFactory class to create Pandas'
DataFrame or Snowflake DataFrame depending on context.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import os
import logging
from typing import Optional, List, Tuple, Union, Any, Dict

import numpy as np
import pandas as pd
from pandas._typing import (
    Dtype,
    Axes,
)
from pandas import DataFrame as DF

from snowflake.snowpark.dataframe import DataFrame as SDF
from snowflake.snowpark._internal.analyzer.snowflake_plan_node \
    import LogicalPlan
from snowflake.snowpark import Session
from snowflake.snowpark.types import StructType

from snowflake_ai.common import DataConnect
from snowflake_ai.connect import SnowConnect, FileConnect



class DataFrameFactory:
    """
    This class provides an uniform dataframe creation interface to create
    a Pandas DataFrame or Snowflake DataFrame instance depending on the
    input context.

    To create a dataframe, construct an appropriate DataConnect object and
    pass it to this factory class together with the data content, optionally
    supplied with the corresponding schema. Note, you can configure Datasets
    to directly get snowflake or pandas dataframe without using this factory
    class.

    Example 1:
        >>> from snowflake_ai.common import DataFrameFactory
        ...
        ... # create snowflake dataframe with SnowConnect
        >>> sdf = DataFrameFactory.create_df(tbl_name, connect)

    Example 2:
        ... # create snowflake dataframe with SnowConnect
        >>> sdf = DataFrameFactory.create_df('select col from tbl', connect)
        
    Example 3:
        ... # create pandas dataframe with FileConnect
        >>> df = DataFrameFactory.create_df(csv_name, connect)

    Example 4:
        ... # create pandas dataframe
        >>> df = DataFrameFactory.create_df([0, 1, 2], columns=['number'])
    """

    @classmethod
    def create_df(
        cls,
        data: Any, 
        connect : Optional[Union[DataConnect, Session]] = None,
        columns: Optional[
            Union[StructType, Tuple, List[str], Axes, None] 
        ] = None,
        index: Optional[Axes] = None,
        dtype: Optional[Dtype] = None, # type: ignore 
    ) -> Union[SDF, DF]:
        """
        Create a Snowflake specific dataframe or Pandas dataframe depending
        on DataConnect types.

        Args:
            data (Any): input Snowflake table/view name, or sql statement,
                or list, tuple, Pandas dataframe, or LogicalPath
            connect (DataConnect): SnowConnect or FileConnect object
            columns (StructType | List | Tuple | Axes): dataframe schema
            index (Axes) : pandas dataframe index
            dtype (Dtype) : pandas dataframe data type

        Returns:
            DataFrame: Snowflake or Pandas Dataframe
        """
        session = None
        if isinstance(connect, SnowConnect):
            session = connect.get_connection()
            col = columns
            if isinstance(columns, tuple) or isinstance(columns, np.ndarray):
                col = list(columns)
            return cls.create_sdf(data, session, col)  # type: ignore
        
        elif isinstance(connect, FileConnect):
            conn: FileConnect = connect
            if str(data).strip() == "":
                return pd.read_csv(conn.current_connection)
            else:
                f = os.path.join(
                    os.path.dirname(
                        os.path.abspath(str(conn.current_connection))
                    ), 
                    str(data)
                )
                return pd.read_csv(f)
        
        elif isinstance(connect, Session):
            col = columns
            if isinstance(columns, tuple) or isinstance(columns, np.ndarray):
                col = list(columns)
            return cls.create_sdf(data, connect, col)  # type: ignore
        
        else:
            return cls.create_pdf(data, columns, index, dtype) # type: ignore


    @classmethod
    def create_sdf(
        cls,
        data: Any, 
        session: Optional[Session] = None,
        columns: Optional[
            Union[StructType, List[str], None] 
        ] = None
    ) -> SDF:
        """
        Create a snowflake dataframe.

        Args:
            data (Any): input Snowflake table/view name, or sql statement,
                or list, tuple, Pandas dataframe, or LogicalPath
            session (Session): snowflake session
            columns (StructType | List ): dataframe schema

        Returns:
            DataFrame: Snowflake Dataframe
        """

        if session is None and isinstance(data, SDF):
            return data.__copy__()
        
        elif session is None and not isinstance(data, SDF):
            raise ValueError(
                "DataframeFactory.create_sdf(): Creation of Snowflake "\
                "DataFrame requires Snowflake connection session!"
            )

        if session is not None and (
            isinstance(data, List) or isinstance(data, tuple) or
            isinstance(data, DF)
        ):
            if columns is None \
                    or isinstance(columns, StructType) \
                    or isinstance(columns, List):
                return session.create_dataframe(data, columns)
            else:
                raise ValueError(
                    "DataframeFactory.create_sdf(): Creation of Snowflake "\
                    "DataFrame requires related schema or Pandas schema!"
                )
        
        elif session is not None and isinstance(data, str):
            q = data.strip().split()
            if len(q) > 1:
                return session.sql(data)
            else:
                return session.table(data)
        
        elif session is not None and isinstance(data, LogicalPlan):
            return SDF(session, data, False)
        
        elif session is not None and (
                isinstance(data, pd.Series) or isinstance(data, Dict)):
            if columns is None \
                    or isinstance(columns, StructType) \
                    or isinstance(columns, List):
                return session.create_dataframe(DF(data), columns)
            else:
                raise ValueError(
                    "DataframeFactory.create_sdf(): Creation of Snowflake "\
                    "DataFrame failed using Dict data or Pandas Series!"
                )
            
        else:
            raise ValueError(
                "DataframeFactory.create_sdf(): Creation of Snowflake "\
                "DataFrame failed, please check input value and type!"
            )
    

    @classmethod
    def create_pdf(
        cls,
        data: Any = None, 
        columns: Optional[
            Union[Tuple, List[str], Axes, None] 
        ] = None, 
        index: Optional[Axes] = None,
        dtype: Optional[Dtype] = None, # type: ignore 
    ) -> DF:
        """
        Create a Pandas dataframe.

        Args:
            data (Any): input Snowflake dataframe,
                or list, tuple, Dict, Pandas Series or dataframe
            columns (Tuple | List | Axes): dataframe schema
            index (Axes) : dataframe index
            dtype (Dtype) : dataframe data type

        Returns:
            DataFrame: Pandas Dataframe
        """

        logger = logging.getLogger(cls.__name__)
        df = DF(data={}, columns=[])
        
        if isinstance(data, SDF):
            df = data.to_pandas()

        elif isinstance(data, DF):
            df = data.copy()
            
        elif isinstance(data, List) or isinstance(data, tuple) \
                or isinstance(data, Dict) or isinstance(data, pd.Series):
            
            if columns is None or isinstance(columns, np.ndarray) or \
                    isinstance(columns, List) or isinstance(columns, tuple):
                df = cls._create_df(data, index, columns)

        else:
            logger.warn(
                f"DataframeFactory.create_pdf(): Initialization with empty "\
                f"Pandas Dataframe."
            )    
            
        return df
    

    @classmethod
    def _create_df(
        cls,
        data = None,
        index: Union[Axes, None] = None,
        columns: Union[Axes, None] = None,
        dtype: Union[Dtype, None] = None, # type: ignore 
        copy: bool = False,
    ) -> pd.DataFrame:
        rdf = pd.DataFrame()
        logger = logging.getLogger(cls.__name__)
        try:
            rdf = pd.DataFrame(data, index, columns, dtype, copy)
        except Exception as e:
            logger.exception(
                "DataFrameFactory._create_df(): Exception occured when "\
                f"creating local pandas dataframe - {e}!"
            )
        return rdf



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains DataSetup class for setting-up, initialization,
configuration, and tearing-down of data Environment
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import os
import glob
import sys
import importlib.util
import logging
import shutil

from pathlib import Path
from types import ModuleType
from typing import Optional, Dict, Union
from pandas import DataFrame as DF
from datetime import *
from dateutil.relativedelta import *
from pandas.tseries.offsets import BusinessDay

from snowflake_ai.common import ConfigType, ConfigKey
from snowflake_ai.common import DataConnect, AppConfig
from snowflake_ai.connect import FileConnect, \
    SnowConnect, DataFrameFactory



class DataSetup:
    """
    This class sets up, initializes, configures or tears down
    the data Environment

    To use this class, instantiate DataSetup with appropriate DataConnect
    as follows:

        >>> from snowflake_ai.common import FileConnect
        >>> from snowflake_ai.snowpandas import DataSetup
        ... 
        >>> conn: FileConnect = FileConnect()
        >>> setup: DataSetup = DataSetup(conn)
        >>> df = setup.load_data()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    def __init__(
        self, 
        datasetup_key: str,
        connect: DataConnect, 
        data: Optional[Dict[str, DF]] = {},
        app_config: AppConfig = None
    ) -> None:
        self.datasetup_key = datasetup_key
        self._connect = connect
        self._data = data
        self.app_config = app_config

        # setup attributes
        self._init_setup_config()



    def get_script_name(self, app_config: Optional[AppConfig] = None) -> str:
        if not self._script_name and (app_config is not None):
            self.set_script_name(app_config)
        return self._script_name


    def set_script_name(self, app_config: AppConfig) -> str :
        p = os.path.join(app_config.root_path, app_config.script_home)
        sys.path.insert(0, os.path.expanduser(p))

        gk, k = AppConfig.split_group_key(self.datasetup_key)
        config : Dict = app_config.get_all_configs()\
                [ConfigType.DataSetups.value][gk][k]
        
        s : str = config.get(ConfigKey.SCRIPT.value)
        m : str = ''
        if s is not None and s:
            m = s.split('.')[0]
        self._logger.debug(
            f"DataSetup.set_script_name(): Script path [{p}]; "\
            f"Script file [{s}]; Module name [{m}].")
        self._script_name = m
        return m


    def load_module(self, name: Optional[str] = '') -> ModuleType:
        if not name:
            name = self._script_name
        return AppConfig.load_module(name)


    def get_connect(self) -> DataConnect:
        """
        Get the data connection for the setup.

        Returns:
            DataConnect: data connection object for the setup
        """        
        return self._connect
    

    def load_data(self, file_name: Optional[str] = None) -> DF:
        """
        Load pandas dataframe from local-csv file connection.

        Returns:
            DataFrame: Pandas dataframe
        """
        if isinstance(self._connect, FileConnect):
            if (file_name is None) or (file_name.strip() == ""):
                return DataFrameFactory.create_pdf(
                    "", self._connect
                )
            else:
                df: DF = DataFrameFactory.create_pdf(
                    file_name, self._connect
                )
                
                df.columns = df.columns.str.strip().str.lower()\
                    .str.replace(' ', '_')
                return df
        elif isinstance(self._connect, SnowConnect):
            pass
        else:
            raise ValueError(
                "DataSetup.load_data(): Error - Loading File requires "\
                "FileConnect!"
            )


    def _init_setup_config(self):
        conf_d = AppConfig.get_all_configs() if self.app_config is None \
                else self.app_config.get_all_configs()
        _, setup_d = AppConfig.get_group_item_config(
                self.datasetup_key,
                ConfigType.DataSetups.value,
                conf_d
            )
        self.setup_type = setup_d.get(ConfigKey.TYPE.value, "")
        self.type = setup_d.get(ConfigKey.TYPE.value, "")
        self.version = setup_d.get(ConfigKey.VERSION.value, "0.1.0")
        self.domain_env = setup_d.get(ConfigKey.DOMAIN_ENV.value, "dev")
        self.data_connect_ref = setup_d.get(ConfigKey.CONN_DATA.value, "")
        self.script = setup_d.get(ConfigKey.SCRIPT.value, "")
        self.class_name = setup_d.get(ConfigKey.CLASS_NAME.value, "")
        self.init_method = setup_d.get(ConfigKey.INIT.value, "init")
        self.clean_up = setup_d.get(ConfigKey.CLEAN_UP.value, "clean_up")



    def _load_file_data(self):
        pass


    def _load_snowflake_data(self):
        pass


    def get_data(self) -> Dict[str, DF]:
        """
        Get all dataframes for exploration and preparation.

        Returns:
            Dict: dictionary of dataframes
        """
        return self._data


    def clean_data(self, file_name: Optional[str] = None):
        self._data = {}
        self._connect.close_connection()


    @staticmethod
    def clean_up_directory(path:str, delete_dir=False):
        DataSetup._logger.debug(
            f"DataSetup.clean_up_directory(): Current_dir [{Path.cwd()}]")
        
        if os.path.isdir(path) and delete_dir:
                try:
                    shutil.rmtree(path)
                except OSError as e:
                    DataSetup._logger.error(
                        f"DataSetup.clean_up_directory(): Error - "\
                        f"{path} : {e.strerror}!"
                    )
        else:
            for file in glob.glob(path, recursive=True):
                if os.path.isfile(file):
                    try:
                        os.remove(file)
                        DataSetup._logger.debug(
                            f'DataSetup.clean_up_directory(): '\
                            f'[{file}] has been deleted.'
                        )
                    except OSError as e:
                        DataSetup._logger.error(
                            f'DataSetup.clean_up_directory(): Error - '\
                            f'{file} : {e.strerror}!'
                        )


    @staticmethod
    def get_calendar_date(
        month: int,  
        n_mon_days : int, 
        year: Optional[int] = 0, 
        n_years : Optional[int] = 0
    ) -> datetime.date:
        """
        Get n_th calendar date from inputted year and month. Use n_years
        for relative year.

        Parameters:
            month (int): month in the year, e.g, 1 - Jan, 2 - Feb, etc.
            n_mon_days (int): n-th calendar days of the month.
            year (int): year number. if obmitted or 0, current year is used.
            n_years (int): future or past n years based on input year value.

        Returns:
            date: date object from datetime
        """
        base_date : datetime = date.today()
        if year <= 0:
            year = base_date.year
        return datetime(
            year=year + n_years, month=month, day=n_mon_days
        ).date()


    @staticmethod
    def get_buinsess_date(
        month: int,  
        n_biz_days : int,
        year: Optional[int] = 0, 
        n_years : Optional[int] = 0
    ) -> datetime.date:
        """
        Get n_th business date from inputted year and month. For relative year,
        use n_years parameter.

        Parameters:
            month (int): month in the year, e.g, 1 - Jan, 2 - Feb, etc.
            n_biz_days (int): n-th business days of the month.
            year (int): year number. if obmitted or 0, current year is used.
            n_years (int): future or past n years based on input year value.

        Returns:
            date: date object from datetime
        """
        base_date : datetime = date.today()
        if year <= 0:
            year = base_date.year
        d = datetime(year=year+n_years, month=month, day=n_biz_days)
        mon_start = d + relativedelta(months=0, day=1)
        offset = BusinessDay(n=n_biz_days)
        return (mon_start + offset).date()


    @staticmethod
    def get_relative_months_calendar_days_date(
        n_months: int, 
        n_cal_days : Optional[int] = 0,
        base_date: Optional[Union[str, datetime, date]] = None
    ) -> datetime.date:
        """
        Get the date of the n-th calendar days (n_cal_days) of future n_months
        (if n_months is positive) or past n_months (if n_month is negative) 
        calculated based on the base input date.

        Parameters:
            n_months (int): Future n months from the current date if it is 
                positive; 0 if it is for the current month; negative if it
                is for the past months.
            n_cal_days (int): n-th's day of the month
            base_date (str, datetime, date): Optional date as the base for 
                the relative date. default is using today's date

        Returns:
            date: date object of the relative date
        """
        if base_date is None:
            base_date : datetime = date.today()
        if isinstance(base_date, date):
            base_date : datetime = datetime(
                year=base_date.year, 
                month=base_date.month,
                day=base_date.day,
            )
        elif isinstance(base_date, str):
            base_date: datetime = datetime.strptime(base_date, "%Y-%m-%d")

        if n_cal_days != 0 :
            mon_days = base_date + relativedelta(
                months=n_months, day=n_cal_days
            )
        else:
            mon_days = base_date + relativedelta(
                months=n_months
            )
        return mon_days.date()
    

    @staticmethod
    def get_relative_months_business_days_date(
        n_months: int, 
        n_biz_days : int,
        base_date: Optional[Union[str, datetime, date]] = None
    ) -> datetime.date:
        """
        Get the date of the n-th business days (n_biz_days) of future n months
        (n_months if positive) or past n months (if negative) calculated based
        on the base input date.

        Parameters:
            n_months (int): Future n months from the current date if it is 
                positive; 0 if it is for the current month; negative if it
                is for the past months.
            n_cal_days (int): n-th's business date of the month
            base_date (str, datetime, date): Optional date as the base for the 
                relative date. default is using the current today's date.

        Returns:
            date: date object of the relative date
        """
        if base_date is None:
            base_date : datetime = date.today()
        if isinstance(base_date, date):
            base_date : datetime = datetime(
                year=base_date.year, 
                month=base_date.month,
                day=base_date.day,
            )
        elif isinstance(base_date, str):
            base_date: datetime = datetime.strptime(base_date, "%Y-%m-%d")

        mon_start = base_date + relativedelta(months=n_months, day=1)
        offset = BusinessDay(n=n_biz_days)
        return (mon_start + offset).date()
    

    @staticmethod
    def get_relative_month_start_date(
        n_years: Optional[int] = 0,
        n_months: Optional[int] = 0,
        base_date: Optional[Union[str, datetime, date]] = None       
    ) -> datetime.date:
        """
        Get the 1st day of the month of future n years (n_years if positive)
        or past n years (if negative) and n months calculated based on the base
        input date's year and month.

        Parameters:
            n_years (int): Future n years from the current date if it is 
                positive; 0 if it is for the current year; negative if it
                is for the past years.
            n_months (int): Relativee n months from the current month if it is 
                positive; 0 if it is for the current month; negative if it
                is for the earlier months.
            base_date (str, datetime, date): Optional date as the base for the 
                relative date. default is using the current today's date.
        """
        if base_date is None:
            base_date : datetime = date.today()
        if isinstance(base_date, date):
            base_date : datetime = datetime(
                year=base_date.year, 
                month=base_date.month,
                day=base_date.day,
            )
        elif isinstance(base_date, str):
            base_date: datetime = datetime.strptime(base_date, "%Y-%m-%d")
        
        mon_start = base_date + relativedelta(
            years=n_years, months=n_months, day=1
        )
        return mon_start.date()
    

    @staticmethod
    def get_relative_month_end_date(
        n_years: Optional[int] = 0,
        n_months: Optional[int] = 0,
        base_date: Optional[Union[datetime, date]] = None       
    ) -> datetime.date:
        """
        Get the last day of the month of future n years (n_years if positive)
        or past n years (if negative) calculated based on the base input
        date.

        Parameters:
            n_years (int): Future n years from the current date if it is 
                positive; 0 if it is for the current year; negative if it
                is for the past years.
            base_date (str, datetime, date): Optional date as the base for the 
                relative date. default is using the current today's date.
        """
        if base_date is None:
            base_date : datetime = date.today()
        if isinstance(base_date, date):
            base_date : datetime = datetime(
                year=base_date.year, 
                month=base_date.month,
                day=base_date.day,
            )
        elif isinstance(base_date, str):
            base_date: datetime = datetime.strptime(base_date, "%Y-%m-%d")

        year_mon = base_date + relativedelta(years=n_years, months=n_months)
        nx_mon_start : date = DataSetup.get_relative_months_calendar_days_date(
            1, 1, year_mon
        )
        nx_mon_d = datetime(
            year=nx_mon_start.year, 
            month=nx_mon_start.month,
            day=nx_mon_start.day,       
        )
        mon_end : datetime = nx_mon_d + relativedelta(days=-1)
        return mon_end.date()



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains a default base application
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import os
from os.path import exists
import sys
from typing import Optional, Union, Dict
from typing import List, Tuple, Any
import logging
from types import ModuleType

from pandas._typing import (
    Dtype,
    Axes,
)
from pandas import DataFrame as DF

from snowflake.snowpark.dataframe import DataFrame as SDF
from snowflake.snowpark.types import StructType
from snowflake.snowpark import Session

from snowflake_ai.common import AppConfig
from snowflake_ai.common import ConfigManager
from snowflake_ai.common import ConfigKey, ConfigType, AppType
from snowflake_ai.connect import ConnectManager, SnowConnect
from snowflake_ai.connect import DataFrameFactory as DFF
from snowflake_ai.snowpandas import SetupManager, DataSetup, SnowSetup
from snowflake_ai.mlops import FlowContext, Pipeline


class BaseApp:
    """
    This class represents a base application's configurations and 
    its corresponding application. Since it is subclass of AppConfig,
    it has the following config directory bootstrapping precedence:

    1) input custom directory 
    2) .snowflake_ai/conf subdir under current directory 
    3) .snowflake_ai/conf subdir under user_home directory
    4) current directory
    5) home directory
    6) conf subdir under snowflake_ai library installation root dir

    Assuming a base application app_1 under a default group named as
        business_group is configured, this app can be created as:

        >>> from snowflake_ai.apps import BaseApp
        ... 
        ... # initialize application config for app_1
        >>> app = BaseApp("business_group.app_1")
        """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    # app_key : app specific config dictionary
    _base_apps : Dict [str, Dict] = {}


    def __init__(
        self,
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ):
        """
        Create a base app based on specific AppConfig object.

        Args:
            app_key: A string representing the application; it can have 
                the format of <app_group>.<application> ('<', '>' not 
                included)
            config_dir: directory path string for custom config load
            config_file: file name string for custom config load
        """
        self.logger = self._logger
        ac: AppConfig = ConfigManager.get_app_config(
                app_key, config_dir, config_file)
        self.logger.debug(
                f"BaseApp.init(): AppConfig - AppKey [{ac.app_key}]; "\
                f"Version [{ac.version}]; Root_path [{ac.root_path}];"\
                f" Script_home [{ac.script_home}].")
        self.appconf = ac
        self.all_configs = AppConfig.get_all_configs()
        self.app_key = ac.app_key
        self.root_path = ac.root_path
        self.app_config = ac.app_config
        self.app_name = ac.app_name
        self.app_group = ac.app_group
        self.app_short_name = ac.app_short_name
        self.type = ac.type
        self.group_config = ac.group_config
        self.version = ac.version
        self.domain_env = ac.domain_env
        self.app_path = ac.app_path
        self.script_home = ac.script_home
        self.app_base_config = ac.app_base_config
        self.app_connect_refs = ac.app_connect_refs
        self.ml_ops_refs = ac.ml_ops_refs
        self.oauth_connect_configs = ac.oauth_connect_configs
        self.data_connect_configs = ac.data_connect_configs
        self.data_setup_refs = ac.data_setup_refs
        self.ml_pipeline_refs = ac.ml_pipeline_refs
        
        if self.type not in [AppType.Default.value, AppType.Console,
                AppType.Notebook.value, AppType.Streamlit.value]:
            raise TypeError(
                "Base.init(): Application type configuration Error!"
            )
        
        # get default snowflake connection
        self.snow_connect = ConnectManager.create_default_snow_connect(
                self.appconf)
        self.default_setup: SnowSetup = SetupManager.create_default_snow_setup(
                self.appconf, self.snow_connect)
        self.setup_module = self.load_module(self.default_setup.script)
        self.default_context = FlowContext()
        self.app_namespaces = self.get_app_namespaces()
        self.default_context.data["app_namespace"] = self.app_namespaces[0]
        self.default_context.data["app_name"] = self.app_namespaces[1]
        self.default_context.debug = False

        # overridden by children    
        session: Session = self.get_snowflake_session({})
        if session is not None:
            self.default_context.session = session
            self.set_traditional_western_week_policy(session)
            if len(self.ml_pipeline_refs) > 0:
                pipeline_key = self.ml_pipeline_refs[0]
                self.default_pipeline = Pipeline(
                        pipeline_key, self.default_context, self.appconf)
        else:
            self.logger.debug(f"BaseApp.init(): Got session [{session}]")



    def set_traditional_western_week_policy(
        self, 
        session: Optional[Session] = None
    ):
        if session is None:
            session = self.get_snowflake_session()
        session.sql("ALTER SESSION SET WEEK_OF_YEAR_POLICY=1, WEEK_START=7")\
            .collect()


    def create_df(
        self,
        data: Any, 
        columns: Optional[
            Union[StructType, Tuple, List[str], Axes, None] 
        ] = None,
        index: Optional[Axes] = None,
        dtype: Optional[Dtype] = None,
        session: Session = None
    ) -> Union[SDF, DF]:
        if session is None:
            session = self.default_context.session
        
        if session is None:
            return DFF.create_df(data, None, columns, index, dtype)
        else:
            return DFF.create_df(
                data, session, columns, index, dtype
            )


    def get_default_snow_connect(self) -> SnowConnect:
        return self.snow_connect
    

    def load_setup_module(self) -> ModuleType:
        return self.default_setup.load_module()


    def get_default_setup(self) -> DataSetup:
        return self.default_setup


    def get_snowflake_session(self, ctx: Dict = {}) -> Session:
        session = None
        if self.snow_connect is not None:
            session = self.snow_connect.get_connection()
            if session is None:
                session = self.snow_connect.create_user_session(ctx)

        return session


    def get_default_pipeline_config(self, pipeline_key):
        _, d = self.get_group_item_config(
                pipeline_key, ConfigType.MLPipelines.value, self.appconf)
        return d


    def run_default_pipeline(self):
        self.default_pipeline.run()
    

    def get_app_namespaces(self) -> Tuple[str, str]:
        return self.appconf.get_app_namespaces()


    def load_configs(
        self,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ) -> Tuple[str, str, Dict]:
        return self.appconf.load_configs(config_dir, config_file)

    
    def load_default_configs(self) -> Dict:
        return self.appconf.load_default_configs()


    def get_group_item_config(
        self,
        group_item_key: str,
        root_key: str,
        configs: Optional[Union[Dict, None]] = None
    ) -> Tuple[str, Dict]:
        return self.appconf.get_group_item_config(
                group_item_key, root_key, configs)


    def load_module(self, script_name: str) -> ModuleType:
        return AppConfig.load_module(script_name)


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains StreamlitApp class representing an application
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import os
from os.path import exists
import sys
from typing import Optional, Union, Dict, Callable
import logging
import toml
from functools import wraps
import streamlit as st
from types import ModuleType
from snowflake.snowpark import Session

from snowflake_ai.common import AppConfig, AppType, ConfigType
from snowflake_ai.apps import AppPage, BaseApp
from snowflake_ai.common import OAuthConnect
from snowflake_ai.connect import AuthCodeConnect, ConnectManager,\
        SnowConnect
from snowflake_ai.snowpandas import SetupManager, DataSetup



class StreamlitApp(BaseApp):
    """
    This class represents a Steamlit application's configurations and 
    its corresponding application. Since it is subclass of AppConfig,
    it has the following config directory bootstrapping precedence:

    1) input custom directory 
    2) snowflake_ai/conf subdir under current directory 
    3) snowflake_ai/conf subdir under user_home directory
    4) current directory
    5) home directory
    6) conf subdir under snowflake_ai library installation root dir

    Assuming a streamlit application app_1 under a default group named as
        streamlit_default is configured, this app can be created as:

        >>> from snowflake_ai.apps import StreamlitApp
        ... 
        ... # initialize application config for app_1
        >>> app = StreamlitApp("streamlit_default.app_1")
        ...
        ... # optionally the configs may be loaded from your custom dir
        ... # app = AppConfig("group_0.app_1", "custom dir", "config.toml")
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    K_PAGE = "page"
    K_PRE_PAGE = "previous_page"
    T_ST_APP = AppType.Streamlit.value


    def __init__(
        self,
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ):
        """
        Creat app specific AppConfig object.

        Args:
            app_key: A string representing the application; it can have 
                the format of <app_group>.<application> ('<', '>' not 
                included)
            config_dir: directory path string for custom config load
            config_file: file name string for custom config load
        """
        super().__init__(app_key, config_dir, config_file)
        self.logger = AppConfig._logger
        if StreamlitApp.T_ST_APP != self.type:
            raise TypeError(
                "StreamlitApp.init(): Application type configuration Error."
            )
        self._setup_streamlit_config()
        self.pages : Dict[str, AppPage] = {}
        self.is_logged_in = False
        self.snow_connect = ConnectManager.create_default_snow_connect(self)
        self.default_setup = SetupManager.create_default_snow_setup(
            self, self.snow_connect
        )
        self.default_setup.load_module()


    def _setup_streamlit_config(
        self, 
        conf_file: Optional[str] = None,
        file_content: Optional[str] = None
    ) -> str:
        """
        Setup streamlit app config file.

        Args:
            conf_file (str): config file name
            file_content (str): custom config file content
        
        Retrun:
            str: streamlit configuration file path
        """
        app_dir = os.path.join(self.root_path, self.script_home)
        p = os.path.join(app_dir, ".streamlit/")
        if not exists(p):
            try:
                os.makedirs(p, exist_ok=True)
            except Exception as e:
                raise ValueError(
                    f"Streamlit._setup_streamlit_dir(): Error [{e}] - "\
                    "Cannot create streamlit configuration directory!"
                )

        if conf_file is None:
            conf_file = "config.toml"
        
        conf_path = os.path.join(p, conf_file)

        if not exists(conf_path):
            if file_content is None:
                file_content = self.get_streamlit_config()
            
            if file_content:
                with open(conf_path, "w") as f:
                    f.write(file_content)
        else:
            with open(conf_path, "r") as f:
                current_content = f.read()
            
            if file_content is None:
                file_content = self.get_streamlit_config()
            
            if current_content != file_content:
                if file_content:
                    with open(conf_path, "w") as f:
                        f.write(file_content)

        return conf_path


    def get_streamlit_config(self) :
        """
        Get streamlit app configuration from config file
        """
        # split app_key in form of group_key.app_key into two variables
        gk, k = AppConfig.split_group_key(self.app_key)
        config: Dict[str, Dict] = self.get_all_configs().get(
            ConfigType.Streamlits.value, {}
        )
        self.logger.debug(
            f"StreamlitApp.get_streamlit_config(): Group_key [{gk}]; App_"\
            f"key [{k}]; Config [{config}]."
        )
        if config:
            config = config.get(gk, {})
            if config:
                config = config.get(k, {})
        
        return toml.dumps(config)
    

    def add_page(self, page:AppPage):
        if self.pages:
            self.pages[page.page_id] = page
        

    def _add_page(
        self,
        page_id: str,
        title: Optional[str] = None,
        icon: Optional[str] = None,
        layout: Optional[str] = "wide",
        show_sidebar: Optional[bool] = True,
        func: Optional[Callable] = None,
    ):
        page = AppPage(
            page_id=page_id,
            title=title,
            icon=icon,
            layout=layout,
            show_sidebar=show_sidebar,
            func=func,
        )
        self.pages[page_id] = page


    def register_page(
        self,
        page_id: str,
        title: Optional[str] = None,
        icon: Optional[str] = None,
        layout: Optional[str] = "wide",
        show_sidebar: Optional[bool] = True,
    ) -> Callable:
        def decorator(func: Callable) -> Callable:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            self._add_page(
                page_id=page_id,
                title=title,
                icon=icon,
                layout=layout,
                show_sidebar=show_sidebar,
                func=wrapper,
            )
            return wrapper

        return decorator
    

    def set_pages_config(self):
        for _, pg in self.pages.items():
            st.set_page_config(
                page_title = pg.title,
                page_icon = pg.icon ,
                layout = pg.layout,
            )


    @staticmethod
    def get_current_page():
        query_params = st.experimental_get_query_params()
        return query_params.get(StreamlitApp.K_PAGE, ["/"])[0]
    

    def navigate_to(self, page_id: str):
        previous_page_id = self.get_current_page()
        st.experimental_set_query_params(
            previous_page = previous_page_id
        )
        st.experimental_set_query_params(page=page_id)
        self.run()


    def run(self):
        current_page_id = self.get_current_page()
        page: AppPage = self.pages.get(current_page_id)
        if page:
            page.func()
        else:
            st.error("Page not found!")


    def request_auth_code(
        self,
        oc:OAuthConnect
    ) -> str:
        oc = ConnectManager.create_default_oauth_connect(self)
        _, cc = AuthCodeConnect.generate_pkce_pair()
        params = {
            AuthCodeConnect.K_CODE_CHALLENGE: cc
        }
        return oc.authorize_request(params)


    def request_access_token(
        self,
        oc: OAuthConnect,
        ap: AppPage = None
    ) -> Dict :
        ok: bool = False
        params = oc.prepare_grant_request()
        auth_cd = params.get("auth_code")
        ctx = {}
        if auth_cd is not None and auth_cd:
            tok_res = oc.grant_request(params)
            self.logger.debug(
                f"StreamlitApp.request_access_token(): "\
                f"Token result - [{tok_res}]"
            )
            print()
            if tok_res:
                dc = ConnectManager.create_default_snow_connect(self)
                ctx = oc.decode_token(
                    tok_res, ["access_token", "refresh_token"]
                )
                self.logger.debug(
                    f"StreamlitApp.request_access_token(): "\
                    f"Decode token, Context [{ctx.items()}]"
                )                
                session = dc.create_session(ctx)
                if ap is not None:
                    ap.session = session
                    self.logger.debug(
                        f"StreamlitApp.request_access_token(): "\
                        f"Creation of snowflake session [OK]"
                    )
                tok: Dict = ctx.get("decoded_access_token")
                if tok:
                    fnm = tok.get("given_name", "")
                    lnm = tok.get("family_name", "")
                    st.sidebar.write(
                        f"Welcome, {fnm} {lnm}!"
                    )
                ok = True
        if not ok:
            st.error("Authorization failed.")

        return ctx


    def request_refresh_token(
        self,
        ctx: Dict,
        oc: OAuthConnect,
        ap: AppPage = None
    ) -> Dict :
        ok: bool = False
        params: Dict = oc.prepare_token_refresh(ctx) 
        refresh_tok = params.get("refresh_token")
        ctx = {}
        if  refresh_tok is not None and refresh_tok:
            tok_res = oc.refresh_token_request(params)
            self.logger.debug(
                f"StreamlitApp.request_refresh_token(): "\
                f"Refersh token result- [{tok_res}]"
            )              
            if tok_res:
                dc = ConnectManager.create_default_snow_connect(self)
                ctx = oc.decode_token(
                    tok_res, ["access_token", "refresh_token"]
                )
                if ctx.get("access_token"):
                    self.logger.debug(
                        f"StreamlitApp.request_refresh_token(): "\
                        f"Get access token [OK]"
                    )
                session = dc.create_session(ctx)
                if ap is not None:
                    ap.session = session
                    self.logger.debug(
                        f"StreamlitApp.request_refresh_token(): "\
                        f"Creation snowflake session [OK]"
                    )

                tok: Dict = ctx.get("decoded_access_token")
                if tok:
                    fnm = tok.get("given_name", "")
                    lnm = tok.get("family_name", "")
                    st.sidebar.write(
                        f"Welcome, {fnm} {lnm}!"
                    )
                ok = True
        if not ok:
            st.error("Authorization failed.")

        return ctx
    


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains SnowEDA and SnowPrep class specific for 

"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


from typing import (
    Dict, Union, Optional, List
)

from pandas import DataFrame as DF
import json
from sklearn.preprocessing import LabelEncoder
from snowflake.snowpark.functions import udf
from snowflake.snowpark.types import IntegerType

from snowflake.snowpark import DataFrame as SDF
from snowflake.snowpark.functions import col, object_construct, CaseExpr
from snowflake.snowpark.functions import lit, array_agg, to_varchar, when
from snowflake.snowpark.functions import sql_expr, iff

from snowflake_ai.snowpandas import EDA, DataPrep, SnowSetup



class SnowEDA(EDA):
    """
    This class extends EDA for Snowflake specific data exploration.

    In general, the framework can bootstrap and use this class implicitly,
    but if the client want to use this class directly, instantiate SnowEDA
    with SnowEDA as follows:

        >>> from snowflake_ai.common import SnowConnect
        >>> from snowflake_ai.snowpandas import SnowSetup
        ...
        >>> conn: SnowConnect = SnowConnect()
        >>> setup: SnowSetup = SnowSetup(conn)
        >>> eda: SnowEDA = SnowEDA(setup)
    
    """
    def __init__(
        self, 
        setup: SnowSetup,
        dfs: Optional[Dict[str, SDF]] = None
    ):
        super().__init__(setup)
        self._sdf = SDF()
        if dfs is not None and isinstance(dfs, Dict):
            for _, value in dfs.items():
                if not isinstance(value, SDF):
                    raise TypeError(
                        f"Dictionary values must be DataFrames, "
                        f"but got {type(value)} instead!"
                    )
            self._data = dfs


    def all_dfs(self) -> Dict[str, SDF]:
        """
        Get all dataframes for exploration.

        Returns:
            Dict: dictionary of dataframes
        """
        return self._data
    

    def active_df(
        self, 
        df_name: Optional[str] = None,
        df: Optional[SDF] = None
    ) -> SDF:
        """
        Get active Snowflake dataframe for exploration.

        Args:
            df_name (str): Snowflake Dataframe name.
            df (DataFrame): set the Snowflake dataframe as active

        Returns:
            DataFrame: active snowflake dataframe for exploration
        """
        if df_name is not None :
            try:
                if df_name in self._data.keys():
                    self._sdf = self._data[df_name]
            except:
                raise ValueError(
                    f"SnowPrep.active_df (): Dictionary of DataFrame doesn't "\
                    f"have the key [{df_name}]!"
                )
            if df is not None:
                self._sdf = df
                self._data[df_name] = df
        elif df is not None:
            self._sdf = df
            self._data[EDA.T_DEFAULT_DF] = df            
        return self._sdf
    

    @staticmethod
    def top_df(df: SDF, n: int = 10, by_col: str = None,
        is_asc: bool = False
    ) -> DF:
        """
        Return pandas dataframe of top n rows from snowflake dataframe
        ordered by one column

        Args:
            df (DataFrame): Snowflake dataframe
            n (int): top n rows
            by_col (str): name of the column ordered by
            is_asc (bool): True if it is ascending order; default False
         
        Returns:
            DataFrame: Pandas Dataframe.
        """
        if by_col is None:
            return df.limit(n).to_pandas()
        else:
            sorted_sdf = df.sort(col(by_col), ascending=is_asc)
            return sorted_sdf.limit(n).to_pandas()
        

    def desc(self, df_name: Optional[str] = None):
        """
        Describe the dataframe by dataframe name.

        Args:
            df_name (str): dataframe name. if emptry, the current
                active dataframe desc is called.
         
        Returns:
            Series or DataFrame: Summary statistics of the Series or
                Dataframe provided.
        """
        if df_name is None:
            return self._df.describe()
        elif df_name is not None and isinstance(self._data, Dict):
            return self._data[df_name].describe()
        else:
            raise TypeError(
                f"SnowPrep.desc(): df_name or initialization of dataframe"\
                f" is required!"
            )



class SnowPrep(DataPrep):
    """
    This class extends DataPrep for Snowflake specific data preparation.

    To use this class directly, instantiate SnowPrep as follows:

        >>> from snowflake_ai.common import SnowConnect
        >>> from snowflake_ai.snowpandas import SnowSetup
        ... 
        >>> conn: SnowConnect = SnowConnect()
        >>> setup: SnowSetup = SnowSetup(conn)
        >>> sp: SnowPrep = SnowPrep(setup)
    
    """

    def __init__(
        self, 
        setup: SnowSetup,
        dfs: Optional[Union[SDF, Dict[str, SDF]]] = None
    ):
        super().__init__(setup)

    
    @property
    def data_setup(self) -> SnowSetup:
        if isinstance(self.data_setup, SnowSetup):
            return self.data_setup
        else:
            raise TypeError(
                "SnowPrep.data_setup(): The datasetup should be SnowSetup!"
            )    


    @staticmethod
    def check_input_columns(df:SDF, input_columns):
        """
        Checks and formats the input_columns parameter.

        If input_columns is None, it defaults to using all columns 
        in the DataFrame. If input_columns is not a list (i.e., it's a
        single column name), it wraps it in a list.

        Parameters:
            df (DataFrame): The DataFrame whose columns are to be checked.
            input_columns (None, str, list): The columns to check. This 
            could be None, a single column name, or a list of column names.

        Returns:
            list: A list of column names from the DataFrame.
        """
        if not input_columns:
            input_columns = df.columns
        else:
            if not isinstance(input_columns, list):
                input_columns = [input_columns]

        return input_columns


    @staticmethod
    def get_columns_unique_categories(
        df: SDF, 
        categories : Dict = {}, 
        cat_cols: List = []
    ):
        """
        Generate a dictionary containing unique categories for each specified
        column in the DataFrame.
        
        Parameters:
        df (DataFrame): The DataFrame to extract categories from.
        categories (str, dict): Parameter to determine whether to automatically 
            generate categories. If empty, unique categories are computed for each
            column. If it's a dictionary, it's assumed that the unique categories 
            for each column are already defined.
        cat_cols (list): The list of column names to extract categories from.
        
        Returns:
            dict: A dictionary where keys are column names and values are lists of
            unique categories.
        """
        if categories == "":
            obj_const_args = []
            
            for c in cat_cols:
                # Get a sorted list of unique string values in the column
                unique_values_agg = array_agg(to_varchar(c), is_distinct=True)\
                    .within_group(to_varchar(col(c)).asc())
                # Add column name and unique values to the argument list
                obj_const_args.extend([lit(c), unique_values_agg])
            
            # Create a DataFrame with one row and one column named "cats",
            # containing a dictionary of unique categories for each column
            df_cats = df.select(
                    object_construct(*obj_const_args).alias("cats")
                )

            cats_dict = json.loads(df_cats.collect()[0]["cats"])
        else: 
            cats_dict = categories

        return cats_dict




# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains a default base application
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import os
from os.path import exists
import sys
from typing import Optional, Union, Dict
from typing import List, Tuple, Any
import logging
from types import ModuleType

from pandas._typing import (
    Dtype,
    Axes,
)
from pandas import DataFrame as DF

from snowflake.snowpark.dataframe import DataFrame as SDF
from snowflake.snowpark.types import StructType
from snowflake.snowpark import Session

from snowflake_ai.apps import BaseApp
from snowflake_ai.connect import DeviceCodeConnect



class ConsoleApp(BaseApp):
    """
    This class represents a console application's configurations and 
    its corresponding application. Since it is subclass of AppConfig,
    it has the following config directory bootstrapping precedence:

    1) input custom directory 
    2) .snowflake_ai/conf subdir under current directory 
    3) .snowflake_ai/conf subdir under user_home directory
    4) current directory
    5) home directory
    6) conf subdir under snowflake_ai library installation root dir

    Assuming a base application app_1 under a default group named as
        business_group is configured, this app can be created as:

        >>> from snowflake_ai.apps import BaseApp
        ... 
        ... # initialize application config for app_1
        >>> app = BaseApp("business_group.app_1")
        """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))


    def __init__(
        self,
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ):
        """
        Creat console app based on specific AppConfig object.

        Args:
            app_key: A string representing the application; it can have 
                the format of <app_group>.<application> ('<', '>' not 
                included)
            config_dir: directory path string for custom config load
            config_file: file name string for custom config load
        """
        super().__init__(app_key, config_dir, config_file)
        self.logger = self._logger



    def get_snowflake_session(self, ctx: Dict = {}) -> Session:
        session = None
        if self.snow_connect is not None:
            session = self.snow_connect.get_connection()
            if session is None:
                dc: DeviceCodeConnect = self.snow_connect.get_oauth_connect()
                self.logger.debug(
                    f"ConsoleApp.get_snowflake_session(): "\
                    f"Auth_type[{dc.auth_type}]"
                )
                res_d = dc.authorize_request()
                res_t = dc.process_authorize_response(res_d)
                ctx = dc.decode_token(
                    res_t, ["access_token", "refresh_token"]
                )
                self.logger.debug(
                    f"ConsoleApp.get_snowflake_session(): "\
                    f"Auth_type[{dc.auth_type}]"
                )
                session = self.snow_connect.create_user_session(ctx)

        return session


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains OAuthConnect implementation of authorization
code flow.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import base64
import hashlib
import os
import requests
import sys
import re
from typing import Optional, Dict, Tuple, List
import logging
import streamlit as st
from urllib.parse import urlencode

from snowflake_ai.common import AppConfig, OAuthConnect



class AuthCodeConnect(OAuthConnect):
    """
    This class represents an application connect for 
    OAuth Authorization code grant flow.

    To use this class, create an instance as following :

        >>> from snowflake_ai.connect import AuthCodeConnect
        ...
        >>> connect = AuthCodeConnect()
        >>> connect.authorize_request()
        >>> connect.grant_request()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))


    K_CODE_VERIFIER = "code_verifier"
    K_CODE_CHALLENGE = "code_challenge"
    K_AUTH_CODE = "auth_code"
    T_CODE = "code"
    T_ACCESS_TOKEN = "access_token"
    T_REFRESH_TOKEN = "refresh_token"


    _initialized = False
    _init_connect_params = {}


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = AuthCodeConnect._logger


    @st.cache_data
    def generate_pkce_pair() -> Tuple[str, str]:
        code_verifier = base64.urlsafe_b64encode(os.urandom(32))\
                .rstrip(b"=").decode("utf-8")
        code_challenge = hashlib.sha256(code_verifier.encode("utf-8"))\
                .digest()
        code_challenge = base64.urlsafe_b64encode(code_challenge)\
                .rstrip(b"=").decode("utf-8")
        return code_verifier, code_challenge


    def authorize_request(self, add_params: Dict) -> str:
        """
        Construct an authorization request url. It is overridden to add
        additional params value for authcode flow.

        Args:
            add_params (Dict): dictionary contains code challenge for PKCE 
                (Proof Key for Code Exchange) 

        Returns:
            str: url string.
        """ 
        if (not add_params) or (not add_params.get(self.K_CODE_CHALLENGE)):
            raise ValueError(
                "AuthcodeConnect.authorize_request(): Additional params dict"\
                " doesn't contain code challenge!"
            )

        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.auth_request_url
        )

        params = {
            "client_id": self.client_id,
            "response_type": self.auth_response_type,
            "redirect_uri": self.redirect_uri,
            "response_mode": self.auth_response_mode,
            "scope": self.scope,
            "code_challenge": add_params[self.K_CODE_CHALLENGE],
            "code_challenge_method": self.code_challenge_method,
            "state": add_params[self.K_CODE_CHALLENGE]
        }

        query_string = urlencode(params)
        auth_url = f"{url}?{query_string}"
        self.logger.debug(
            f"AuthcodeConnect.authorize_request(): Auth_url - [{auth_url}]"
        )
        return auth_url
    

    def prepare_grant_request(self, add_params={}) -> Dict:
        """
        After redirect, prepare parameters for grant token request.
        """
        query_string = st.experimental_get_query_params()
        code: List = query_string.get(self.T_CODE)
        code_verifier, _ = self.generate_pkce_pair()

        add_params = {
            AuthCodeConnect.K_AUTH_CODE: code[0],
            AuthCodeConnect.K_CODE_VERIFIER: code_verifier,
            AuthCodeConnect.K_CLIENT_SECRET: self.client_secret_env
        }
        return add_params


    def grant_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get access token. This method should be
        overwritten by its child class.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.grant_token_request_url
        )

        code = add_params.get(self.K_AUTH_CODE)
        code_verifier = add_params.get(self.K_CODE_VERIFIER)
        sec_env = add_params.get(self.K_CLIENT_SECRET)
        secret = os.environ[str(sec_env)]

        if not code or not code_verifier:
            raise ValueError(
                "AuthCodeConnect.grant_request(): Missing required params "\
                "- code or code_verifier!"
        )
        payload = {
            "client_id": self.client_id,
            "client_secret": secret,
            "scope": self.scope,
            "grant_type": self.grant_type,
            "code": code,
            "redirect_uri": self.redirect_uri,
            "code_verifier": code_verifier,
            "state" : "loggedin"
        }

        headers = {"Content-Type": self.content_type}
        self.logger.debug(
            f"AuthCodeConnect.grant_request(): Payload => {payload}"
        )

        try:
            response = requests.post(url, data=payload, headers=headers)
            response.raise_for_status()
            return response.json()
        except requests.HTTPError as http_err:
            self.logger.error(
                f"AuthCodeConnect.grant_request(): HTTP error occurred:"\
                f" {http_err}!"
            )
            return {}
        except Exception as err:
            self.logger.error(
                f"AuthCodeConnect.grant_request(): Other error occurred:"\
                f" {err}!"
            )
            return {}


    def prepare_token_refresh(self, ctx: Dict) -> Dict: 
        add_params = {
            AuthCodeConnect.T_REFRESH_TOKEN: ctx["refresh_token"],
            AuthCodeConnect.K_CLIENT_SECRET: self.client_secret_env
        }
        return add_params


    def refresh_token_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get refresh token.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.grant_token_request_url
        )

        refresh_tok = add_params.get(self.T_REFRESH_TOKEN)
        sec_env = add_params.get(self.K_CLIENT_SECRET)
        secret = os.environ[str(sec_env)]

        payload = {
            "client_id": self.client_id,
            "scope": self.scope,
            "grant_type": self.T_REFRESH_TOKEN,
            "refresh_token": refresh_tok,
            "client_secret": secret
        }

        headers = {"Content-Type": self.content_type}
        self.logger.debug(
            f"AuthCodeConnect.refresh_token_request(): Payload => {payload}"
        )

        try:
            response = requests.post(url, data=payload, headers=headers)
            response.raise_for_status()
            return response.json()
        except requests.HTTPError as http_err:
            self.logger.error(
                f"AuthCodeConnect.refresh_token_request(): "\
                f"HTTP error occurred: {http_err}!"
            )
            return {}
        except Exception as err:
            self.logger.error(
                f"AuthCodeConnect.refresh_token_request(): "\
                f"Other error occurred: {err}!"
            )
            return {}


    def setup_connect(self, params: Optional[Dict] = None):
        """
        Setup OAuthConnect with proper parameters initialized; it
        should be extended by its child class.

        Args:
            params (Dict): input parameters for connection setup.

        Returns:
            OAuthConnect: self with proper parameters initialized
        """
        super().setup_connect(params)
        self.redirect_uri = params.get("redirect_uri", '')
        self.code_challenge_method = params.get(
            "code_challenge_method", 'S256'
        )
        self.auth_response_mode = params.get("auth_response_mode", "query")
        return self



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains client credentials flow OAuthConnect class.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import time
import os
import requests
import sys
import re
from typing import Optional, Dict, Tuple, List
import logging
import streamlit as st
from urllib.parse import urlencode

from snowflake_ai.common import AppConfig, OAuthConnect



class ClientCredsConnect(OAuthConnect):
    """
    This class represents anapplication connect for 
    OAuth Client Credentials grant flow.

    To use this class, create an instance as following :

        >>> from snowflake_ai.connect import ClientCredsConnect
        ...
        >>> connect = ClientCredsConnect()
        >>> connect.grant_request()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = ClientCredsConnect._logger



    def authorize_request(self, add_params: Dict = {}) -> Dict:
        return add_params
    

    def process_authorize_response(self, json_ctx: Dict) -> Dict:
        return json_ctx
    

    def prepare_grant_request(self, add_params = {}) -> Dict:
        return add_params


    def grant_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get access token.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.grant_token_request_url
        )
        sec_env = add_params.get(self.K_CLIENT_SECRET)
        secret = os.environ[str(sec_env)]

        payload = {
            "client_id": self.client_id,
            "client_secret": secret,
            "scope": self.scope,
            "grant_type": self.grant_type,
        }

        headers = {"Content-Type": self.content_type}
        self.logger.debug(
            f"ClientCredsConnect.grant_request(): Payload => {payload}"
        )

        try:
            response = requests.post(url, data=payload, headers=headers)
            response.raise_for_status()
            return response.json()
        except requests.HTTPError as http_err:
            self.logger.error(
                f"ClientCredsConnect.grant_request(): HTTP error occurred:"\
                f" {http_err}!"
            )
            return {}
        except Exception as err:
            self.logger.error(
                f"ClientCredsConnect.grant_request(): Other error occurred:"\
                f" {err}!"
            )
            return {}
    

    def prepare_token_refresh(self, ctx: Dict) -> Dict: 
        return {}


    def refresh_token_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get refresh token.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        return {}


    def setup_connect(self, params: Optional[Dict] = None):
        """
        Setup OAuthConnect with proper parameters initialized; it
        should be extended by its child class.

        Args:
            params (Dict): input parameters for connection setup.

        Returns:
            OAuthConnect: self with proper parameters initialized
        """
        super().setup_connect(params)
        return self



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains FileConnect class representing a specific file
connection.

Example:
    To use this class, just instantiate FileConnect as follows:

        from snowflake_ai.common import FileConnect

        connect: FileConnect = FileConnect()
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import os
from os.path import exists
import logging
from typing import Dict, Optional

from snowflake_ai.common import AppConfig, DataConnect



class FileConnect(DataConnect):
    """
    This class provides various types of file connection.

    Example:

    To use this class, just instantiate FileConnect as follows:

        >>> from snowflake_ai.common import FileConnect
        ... 
        >>> connect: FileConnect = FileConnect()
    """

    T_FILE_CONN = AppConfig.T_CONN_FILE
    DEF_FILE_CONN = "file_def"
    DEF_FILE_FMT = "csv"
    DEF_STRG_TYPE = "local"

    _logger = logging.getLogger(__name__)


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = FileConnect._logger
        if self.connect_group == DataConnect.K_DATA_CONN and \
                self.connect_params:
            try:
                self.storage_type = self.connect_params["storage_type"]
                self.format = self.connect_params["format"]
                self.dir_path = self.connect_params["dir_path"]
                self.file_name = self.connect_params["file_name"]
            except KeyError as e:
                raise KeyError(
                    f"FileConnect.init(): Configuration Error - {e}"
                )
        else:
            raise ValueError(
                "FileConnect initialization configuration error"
            )
        if (not self.connect_type) or (
            self.connect_type != AppConfig.T_CONN_FILE
        ):   
            raise ValueError(
                "FileConnect initialization data connect type error"
            )


    def create_connection(self, params: Dict):
        """
        Create a file based connection based on the specific configuration.
        A set of types are supported:
            local-csv, azure-adls, azure-blob

        Args:
            params (Dict): Configuration dictionary as input parameters.

        Returns:
            object: file path or connection object
        """
        if params is None:
            raise ValueError(
                f"FileConnect.create_connection params cannot be None"
            )
        try:
            if self.storage_type == "local":
                if self.format == "csv":
                    self.set_current_connection(self._do_local_csv(params))
            elif self.storage_type == "azure_adls":
                self.set_current_connection(self._do_azure_adls(params))
            elif self.storage_type == "azure_blob":
                self.set_current_connection(self._do_azure_blob(params))
        except Exception as e:
            self.logger.exception(
                f"FileConnect.create_connection cannot create "\
                f"storage_type={self.storage_type} file connect: {e}"
            )
        return self.get_current_connection()


    def get_connection(self, connect_key: Optional[str] = None) -> str:
        """
        Get lazy created file connection. If no input connect_key,
        default file connection is created.

        Args:
            connect_key (str): data connect key.

        Returns:
            object: data connection object, e.g., local csv file path
        """
        conn = super().get_connection(connect_key)
        if self.connect_params["type"] != FileConnect.T_FILE_CONN:
            raise TypeError(
                f"FileConnect.get_connection() type error {connect_key}"
            )
        return conn


    def _do_local_csv(self, params) -> str:
        path :str = params["dir_path"]
        file :str = params["file_name"]
        
        file_conn = ''
        if not file:
            file_conn = os.path.dirname(os.path.abspath(path))
        else:
            file_conn = os.path.join(
                os.path.dirname(os.path.abspath(path)), file
            )

        if exists(file_conn):
            self.set_current_connection(file_conn)
        else:
            raise ValueError(
                f"FileConnect._do_local_csv() Cannot load file from local "\
                f"csv path, Parameter paths value Error: dir_path=>{path},"\
                f" file_name=>{file}"
            )

        return self.get_current_connection()


    def _do_azure_adls(self, params) -> str:
        # TO-DO
        return ''


    def _do_azure_blob(self, params) -> str:
        # TO-DO
        return ''



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains SetupManager class representing a factory class to 
create Setup objects.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import sys
from typing import Optional, Dict, List
import logging

from snowflake_ai.common import AppConfig, ConfigKey, ConfigType
from snowflake_ai.connect import AppConnect, SnowConnect
from snowflake_ai.snowpandas import DataSetup, SnowSetup



class SetupManager:
    """
    This class represents a factory class for creating DataSetup 
    objects.

    To use this class, create instance from AppConfig or extend
    it in a child class for creation of a DataSetup as following:

        >>> from snowflake_ai.snowpandas import SetupManager
        ...
        >>> ac = AppConfig("group_0.app_1")
        >>> setup = SetupManager(ac).create_default_data_setup()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    # whether application has been initialized
    _init_d : Dict[str, bool] = {}

    # application data setup list
    _data_setups_d : Dict[str, List] = {}
    

    @classmethod
    def create_default_snow_setup(
        cls, 
        app_config: AppConfig,
        def_conn: Optional[SnowConnect] = None
    ) -> SnowSetup:
        """
        Create default snowflake data setup.

        Args:
            app_config (AppConfig):
            def_conn (SnowConnect):
            
        Returns:
            SnowSetup: default snowflake data setup
        """   
        _initialized = cls._init_d.get(app_config.app_key)
        cls._logger.debug(
            "SetupManager.create_default_snow_setup(): Initialize "\
            f"App_key [{app_config.app_key}] application setup."
        )
        if _initialized is None or not _initialized:
            cls.create_data_setups(app_config, def_conn)
        for s in cls._data_setups_d[app_config.app_key]:
            if isinstance(s, SnowSetup):
                return s
        return SnowSetup()
    

    @classmethod
    def create_data_setups(
        cls, 
        app_config: AppConfig,
        def_conn: Optional[SnowConnect] = None
    ) -> List:
        """
        Create list of application data setups instances

        Args:
            app_config (AppConfig): Application configuration object
            connect (SnowConnect): Default SNowflake Connect object
            
        Returns:
            List: application data setup instances
        """ 
        if app_config is None:
            raise ValueError(
                "SetupManager.create_data_setups(): AppConfig is required!"
            )
        
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls._data_setups_d[app_config.app_key] = []

            if not app_config.data_setup_refs:
                cls._do_data_setup(app_config.app_key, app_config, def_conn)

            for ck in app_config.data_setup_refs:
                cls._do_data_setup(ck, app_config, def_conn)
            
            cls._init_d[app_config.app_key] = True

        return cls._data_setups_d[app_config.app_key]
    

    @classmethod
    def _do_data_setup(
        cls,
        datasetup_key: str,
        app_config: AppConfig,
        def_conn: Optional[SnowConnect] = None
    )  -> None:
        gk, k = AppConfig.split_group_key(datasetup_key)
        config: Dict = app_config.get_all_configs()\
                [ConfigType.DataSetups.value][gk][k]
                
        if (config[ConfigKey.TYPE.value] == AppConfig.T_CONN_SNFLK)\
                and (config.get(ConfigKey.CONN_DATA.value) is not None):
            dconn_ref = config.get(ConfigKey.CONN_DATA.value)
            ac = AppConnect.get_app_connects().get(dconn_ref)
            if ac is None:
                ac = SnowConnect(dconn_ref, app_config)
                AppConnect.get_app_connects()[dconn_ref] = ac

            if def_conn is None:
                def_conn = ac

            setup : SnowSetup = SnowSetup(datasetup_key, def_conn)
            cls._data_setups_d[app_config.app_key]\
                    .append(setup)
            setup.set_script_name(app_config)

            cls._logger.debug(
                "SetupManager._do_data_setups(): "\
                f"Initialize [{config[ConfigKey.TYPE.value]}] "\
                f"DataSetup with key [{datasetup_key}]."
            )



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains ConfigManager class for managing AppConfig objects
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import logging
from typing import Dict, List, Optional, Union

from snowflake_ai.common import AppConfig


# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s]  %(message)s'
)



class ConfigManager:
    """
    This class manages multiple AppConfig objects to ensure only
    one instance for a particular app_key. 
    """

    _logger = logging.getLogger(__name__)
    _app_confs : Dict[str, AppConfig] = {}


    def __init__(self):
        super().__init__()



    @staticmethod
    def get_app_config(
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ) -> AppConfig:
        if not app_key:
            ConfigManager._logger.error(
                    f"ConfigManager.get_app_config(): Error - "\
                    f"app_key [{app_key}] is empty!")
            return None

        ac = ConfigManager._app_confs.get(app_key)
        if ac is None:
            ac = AppConfig(
                app_key=app_key, 
                config_dir=config_dir, 
                config_file=config_file
            )
            ConfigManager._app_confs[app_key] = ac
        
        return ac



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains FlowContext class representing the context
of MLOps flow
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import logging
from typing import Any, Callable, Dict, List, Optional, Tuple
from snowflake.snowpark import DataFrame as SDF
from snowflake.snowpark import Session


class FlowContext:
    """
    This class provides a context for the MLOps flow consisting
    of direct input (prior processing/task), context input (full
    historical context sequence), current output, and related
    metadata.

    Example:

    To use this class, instantiate the initial context:

        >>> from snowflake_ai.mlops import FlowContext
        ... 
        >>> ctx: FlowContext = FlowContext()
    """

    FLG_DEF = 0
    T_STEP_ITER = "iteration"



    def __init__(self) -> None:
        self.logger = logging.getLogger(__name__)
        self.primary_session: Session = None
        self.secondary_session: Session = None
        self.session: Session = self.primary_session
        self.pipelines: Dict[str, Dict] = {}
        self.data: Dict[str, Any] = {}
        self.indexes: Dict[str, Any] = {}
        self.models: Dict[str, Any] = {}
        self.params: Dict[str, Any] = {}
        self.metrics: Dict[str, Any] = {}
        self.metadata = {}
        self.outputs : Dict[str, Any] = {}
        self.debug_info: Dict[str, Any] = {}
        self.debug = True

        self.pipeline: Dict[str, Callable] = {}
        self.step_types: List[str] = []
        self.loggings = {}
        self.process_flag: int = FlowContext.FLG_DEF
        self.direct_inputs = {}
        self.context_inputs = []
    


    @staticmethod
    def exists_in_snowflake(
            session: Session, 
            db_nm: str,
            sch_nm: str,
            tbl_nm: str
        ) -> bool:
        s, db_name, sch_name = "", "", ""
        if db_nm is not None and db_nm.strip():
            db_name = db_nm.strip()        
        if sch_nm is not None and sch_nm.strip():
            sch_name = sch_nm.strip()
        
        if db_name:
            s = f"""
                SELECT COUNT(*) as CNT
                FROM information_schema.tables
                WHERE table_catalog = '{db_name}' and
                table_schema = '{sch_nm}' and 
                table_name = '{tbl_nm}'
            """
        elif sch_name:
            s = f"""
                SELECT COUNT(*) as CNT
                FROM information_schema.tables
                WHERE table_schema = '{sch_nm}' and 
                table_name = '{tbl_nm}'
            """
        else:
            s = f"""
                SELECT COUNT(*) as CNT
                FROM information_schema.tables
                WHERE table_name = '{tbl_nm}'
            """
        t:SDF = session.sql(s)
        n:int = t.collect()[0]["CNT"]
        return False if n == 0 else True


    def exist(self, table_name:str) -> bool:
        """
        Check whether a fully qualified table exist in Snowflake session.
        Note: table_name would be converted all uppercase.
        """
        nl = table_name.rsplit('.', 2)
        if self.session is None or len(nl) == 0:
            return False
        
        if len(nl) == 1:
            return self.exists_in_snowflake(
                self.session, None, None, nl[0].upper()
            )
        elif len(nl) == 2:
            return self.exists_in_snowflake(
                self.session, None, nl[0].upper(), nl[1].upper()
            )
        else:
            return self.exists_in_snowflake(
                self.session, nl[0].upper(), 
                nl[1].upper(), nl[2].upper()
            )


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains EDA and DataPrep class for general or pandas-like 
data exploration, splits, and preprocessing.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


from typing import (
    Dict, Union, Optional, List, Iterable
)
from pandas import DataFrame, Series, to_numeric

import matplotlib.pyplot as plt

from snowflake_ai.common import DataConnect
from snowflake_ai.snowpandas import DataSetup



class EDA:
    """
    This class provides pandas dataframe exploration.

    To use this class, initialize it with DataSetup as follows:

        >>> from snowflake_ai.snowpandas import DataSetup
        ... 
        >>> eda: EDA = EDA(setup)
        
    """    
    T_DEFAULT_DF = "default"


    def __init__(
        self, 
        setup: DataSetup,
        data: Optional[Dict[str, DataFrame]] = {}
    ):
        self._setup = setup
        self._active_df = DataFrame()
        self._data = data


    @property
    def data_setup(self):
        """
        Get DataPrep's Setup.

        Returns:
            DataSetup: data setup for the data preparation
        """
        return self._setup


    def active_df(
        self, 
        df_name: Optional[str] = None,
        df: Optional[DataFrame] = None
    ) -> DataFrame:
        """
        Get active dataframe for exploration.

        Args:
            df_name (str): Dataframe name.
            df (DataFrame): set the dataframe as active

        Returns:
            DataFrame: active dataframe for exploration
        """
        if df_name is not None :
            try:
                if df_name in self._data.keys():
                    self._active_df = self._data[df_name]
            except:
                raise ValueError(
                    f"Dictionary of DataFrame doesn't have the key: {df_name}"
                )
            if df is not None:
                self._active_df = df
                self._data[df_name] = df
        elif df is not None:
            self._active_df = df
            self._data[EDA.T_DEFAULT_DF] = df            
        return self._active_df


    @staticmethod
    def top_df(df: DataFrame, n: int = 10, by_col: str = None,
        is_asc: bool = False
    ):
        """
        Return top n rows from Pandas dataframe ordered by one column

        Args:
            df (DataFrame): Pandas dataframe
            n (int): top n rows
            by_col (str): name of the column ordered by
            is_asc (bool): True if it is ascending order; default False
         
        Returns:
            Series or DataFrame: Pandas Series or Dataframe.
        """
        if by_col is None:
            return df.head(n)
        else:
            sorted_df = df.sort_values(by=by_col, ascending=is_asc)
            return sorted_df.head(n)


    def top(self, df_name: Optional[str] = None, n: int = 10, 
        by_col: str = None, is_asc: bool = False
    ):
        """
        Return top n rows from current active Pandas dataframe or input 
        dataframe ordered by one column

        Args:
            df_name (str): dataframe name

        Retruns:
            Series or DataFrame: Pandas Series or DataFrame
        """
        return EDA.top_df(self.active_df(df_name), n, by_col, is_asc)    


    @staticmethod
    def desc_df(
        df: DataFrame, 
        includes: Optional[List] = None, 
        excludes: Optional[List] = None
    ):
        """
        Describe the dataframe.

        Args:
            df (DataFrame): dataframe to be described
            includes (List): Includes list of data types in the result
            excludes (List) : Excludes the provided data types from 
                the result. 
        Returns:
            Series or DataFrame: Summary statistics of the Series or
                Dataframe provided.
        """
        if includes is None and excludes is None:
            return df.describe()
        elif includes is not None and excludes is None:
            return df.describe(include = includes)
        elif includes is None and excludes is not None:
            return df.describe(exclude = excludes)
        else:
            return df.describe(include = includes, exclude= excludes)


    def desc(self, df_name: Optional[str] = None):
        """
        Describe the dataframe by dataframe name.

        Args:
            df_name (str): dataframe name. if emptry, the current
                active dataframe desc is called.
         
        Returns:
            Series or DataFrame: Summary statistics of the Series or
                Dataframe provided.
        """
        return EDA.desc(self.active_df(df_name))
    

    @staticmethod
    def corr_df(df: DataFrame, label_name: str):
        """
        Compute correlation of columns against target label.

        Args:
            df (dataframe): input dataframe
            label_name (str): target label name.
         
        Returns:
            Correlation matrix.
        """
        corr_mat = df.corr()
        return corr_mat[label_name].sort_values(ascending=False)


    def corr(self, label_name: str, feat_name: Optional[str] = None) \
            -> Union[float, Series, None]:
        """
        Compute correlation between target label and feature column.

        Args:
            label_name (str): target label.
            feat_name (str): feature column.
         
        Returns:
            Correlation matrix.
        """
        if label_name in self._active_df.columns and feat_name is None:
            corr_mat: DataFrame = self._active_df.corr()
            return corr_mat[label_name].sort_values(ascending=False)
        elif label_name in self._active_df.colums and \
                feat_name in self._active_df.columns:
            self._active_df[feat_name] = to_numeric(
                self._active_df[feat_name], errors='coerce'
            )
            self._active_df[label_name] = to_numeric(
                self._active_df[label_name], errors='coerce')
            return Series(self._active_df[feat_name]).corr(Series(
                self._active_df[label_name])
            )
        else:
            raise ValueError(
                f"Check label name {label_name} or feature name {feat_name} "\
                    "is not part of active dataframe's columns"
            )


    def plot(
        self, feat_name, show_hist: bool = True, \
        label_name: Optional[str] = None
    ):
        """
        Show feature column's histgram and scatter plot if target label
        is specified.

        Args:
            show_hist (bool): show histgram if true.
            label_name (str): target label is supplied show scatter plot.
        """        
        if show_hist:
            self._active_df[feat_name].hist()
            plt.show()
        if label_name is not None:
            plt.scatter(self._data[feat_name], self._data[label_name])
            plt.show()



class DataPrep:
    """
    This class provides general data preparation

    To use this class, initialize it with DataSetup as follows:

        from snowflake_ai.snowpandas import DataSetup

        prep: DataPrep = DataPrep(setup)
        
    """
    def __init__(
        self, 
        setup: DataSetup,
        data: Optional[Dict[str, DataFrame]] = {}
    ):
        self._setup = setup
        self._data = data 


    def process(self, *args) -> Iterable[DataFrame]:
        df: DataFrame = DataFrame()
        # subclass specific processing logic
        yield df


    def end_partition(self) -> Iterable[Dict[str, DataFrame]]:
        yield self._data


    def get_setup(self) -> DataSetup:
        return self._setup


    @property
    def data_setup(self) -> DataSetup:
        return self._setup
    



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains MLOps class for ML flow operationization
targeting production domain/environment
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import logging
from concurrent.futures import ThreadPoolExecutor
from typing import Dict, List

from snowflake_ai.common import AppConfig
from snowflake_ai.mlops import FlowContext
from snowflake_ai.mlops import Pipeline, TaskType


# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s]  %(message)s'
)



class MLOps:
    pass



class PipelineFlow:
    """
    This class manages multiple pipelines as an overall flow in MLOps
    operationalizing the model pipeline built, including deloyment
    serving the model, monitor and re-train the model 
    (To-DO).
    """

    def __init__(
            self,
            flow_key: str,
            ctx: FlowContext = None,
            app_config: AppConfig = None            
    ) -> None:
        super().__init__()
        self.logger = logging.getLogger(__name__)


    def execute(self):
        pass



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains SnowConnect class representing a specific Snowflake
connection.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import os
import sys
import logging
import types
from typing import List, Dict, Iterator, Optional

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization

from pandas import DataFrame as DF

from snowflake.snowpark import Session
from snowflake.snowpark import Row
from snowflake.snowpark import DataFrame as SDF

from snowflake_ai.common import AppConfig, ConfigKey
from snowflake_ai.common import AppConnect, DataConnect
from snowflake_ai.connect import AuthCodeConnect, DeviceCodeConnect
from snowflake_ai.connect import ClientCredsConnect



class SnowConnect(DataConnect):
    """
    This class provides specific Snowflake connection.

    Example:

    To use this class, just instantiate SnowConnect as follows:

        >>> from snowflake_ai.common import SnowConnect
        ... 
        >>> connect: SnowConnect = SnowConnect()
        >>> session = connect.get_session()
    """

    T_SNOWFLAKE_CONN = AppConnect.T_SNOWFLAKE_CONN
    T_AUTH_SNOWFLAKE = AppConnect.T_AUTH_SNOWFLAKE
    T_AUTH_KEYPAIR = AppConnect.T_AUTH_KEYPAIR
    T_AUTH_EXT_BROWSER = AppConnect.T_AUTH_EXT_BROWSER
    T_AUTH_OAUTH = AppConnect.T_AUTH_OAUTH

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = SnowConnect._logger

        if not (self.connect_group == DataConnect.K_DATA_CONN and \
                self.connect_params):
            self.logger.error(
                f"SnowConnect.init(): ConnectType [{self.connect_type}];"\
                f"ConnectGroup [{self.connect_group}]; ConnectName ["\
                f"{self.connect_name}]"
            )
            raise ValueError(
                "SnowConnect.init(): SnowConnect configuration Error!"
            )
        if (not self.connect_type) or (
                self.connect_type != AppConfig.T_CONN_SNFLK):
            raise ValueError(
                f"SnowConnect.init(): Error - Type [{self.connect_type}]!"
            )
        
        # setup referenced oauth connect
        self.logger.debug(
            f"SnowConnect.init(): Oauth_connect_ref[{self.oauth_connect_ref}]"
        )
        if self.oauth_connect_ref:
            ot = self.oauth_flow_type if self.oauth_flow_type else \
                    AppConfig.T_OAUTH_DEVICE
            self.oauth_connect = None
            conn = AppConnect._connects.get(self.oauth_connect_ref)
            if conn is None:
                if ot == AppConfig.T_OAUTH_DEVICE:
                    self.oauth_connect = DeviceCodeConnect(
                            self.oauth_connect_ref,
                            self.app_config
                        )

                elif ot == AppConfig.T_OAUTH_CODE:
                    self.oauth_connect = AuthCodeConnect(
                            self.oauth_connect_ref,
                            self.app_config
                        ) 
                 
                elif ot == AppConfig.T_OAUTH_CREDS:
                    self.oauth_connect = ClientCredsConnect(
                            self.oauth_connect_ref,
                            self.app_config
                        )               

                AppConnect._connects[self.oauth_connect_ref] = \
                        self.oauth_connect
            else:
                self.oauth_connect = conn

        # create data service connection and store at DataConnect class
        if (self.connect_key is not None) and self.connect_key:
            if ((self.data_connections.get(self.connect_key) is None) \
                    or (not self.is_current_active())) \
                    and (self.is_service_connect()):
                self.data_connections[self.connect_key] = \
                    self.create_connection(self.connect_params)
                self.set_current_connection(
                    self.data_connections[self.connect_key]
                )

        self.logger.debug(f"SnowConnect.init(): Connect_key"\
                f" [{self.connect_key}]; DataConnections => "\
                f"{self.data_connections}; Is_service_connect ["\
                f"{self.is_service_connect()}]")

        # setup initialization
        self.init_connects()



    def get_oauth_connect(self):
        return self.oauth_connect
    

    def create_connection(self, params: Dict):
        """
        Create a snowflake service connection based on the specific 
        configuration. A set of authentication types are supported:
            snowflake, keypair, oauth and externalbrowser
        Note: user specific connection would be in form of session which
        is typically created through oauth connect.

        Args:
            params (Dict): Configuration dictionary as input parameters.

        Returns:
            object: snowflake connection session
        """
        if params is None:
            raise ValueError(
                "SnowConnect.create_connection(): Input params cannot be None!"
            )
        session = None
        try:
            auth = params[ConfigKey.AUTH_TYPE.value]
            self.logger.debug(
                "SnowConnect.create_connection(): Try to "\
                f"connect to snowflake using [{auth}] auth."
            )
            if auth == AppConnect.T_AUTH_SNOWFLAKE:
                session = self._do_snowflake_auth(params)

            elif auth == AppConnect.T_AUTH_KEYPAIR:
                session = self._do_keypair_auth(params)

            elif auth == AppConnect.T_AUTH_EXT_BROWSER: 
                # should not be used to create svc connection
                self.logger.warning(
                    "SnowConnect.create_connection(): Use create_session "\
                    "to create externalbrowser enabled connection."
                )
            elif auth == AppConnect.T_AUTH_OAUTH:
                if self.oauth_flow_type == AppConfig.T_OAUTH_CREDS:
                    if self.oauth_connect is None:
                        self.logger.error(
                            "SnowConnect.create_connection(): OAuth "\
                            f"[{self.connect_key}] connect is None!"
                        )                        
                        return None
                    cc: ClientCredsConnect = self.oauth_connect
                    d_rs: Dict = cc.grant_request(self.oauth_connect_config)
                    ctx: Dict = cc.decode_token(d_rs, ["access_token"])
                    self.logger.debug(
                        "SnowConnect.create_connection(): OAuth response "\
                        f"decoded_token => {ctx}."
                    )
                    session = self._do_oauth(self.connect_params, ctx)
                else:
                    self.logger.warning(
                        "SnowConnect.create_connection(): No action in "\
                        "create_session to create oauth enabled connection!"
                    )
        except Exception as e:
            self.logger.exception(
                f"SnowConnect.create_connection(): Cannot use "\
                f"auth_type [{auth}] to creaet snowflake session - {e}!"
            )
        return session


    def get_connection(self, connect_key: Optional[str] = None):
        """
        Get lazy created snowflake shared connection.

        Args:
            connect_key (str): data connect key in form of
                "data_connects".<connect_name>

        Returns:
            object: snowflake connection object, i.e., snowflake session
        """
        conn = super().get_connection(connect_key)
        if conn is not None and (
            (self.connect_params[ConfigKey.TYPE.value] != \
                    SnowConnect.T_SNOWFLAKE_CONN) or 
                    (not isinstance(conn, Session))
        ):
            raise TypeError(
                f"SnowConnect.get_connection(): Type Error - {connect_key}!"
            )
        elif conn is None:
            self.logger.warning(
                "SnowConnect.get_connection():  Shared/Service connection "\
                f"is None; Current_connect_key [{connect_key}]; "
                f"Current_connect_type [{self.connect_params['type']}]."
            )

        return conn


    @classmethod
    def is_session_active(cls, session: Session) -> bool:
        rb = False
        try:
            r = cls.dql(session, "select current_role()")           
            try:
                pd = next(r)
                cls._logger.debug(
                    "SnowConnect.is_session_active(): Active "
                    f"session using role [{pd}]"
                )
                if DF(pd).shape[0] > 0:
                    rb = True
            except Exception as ee:
                cls._logger.warning(
                    "SnowConnect.is_session_active(): No Result "\
                    f"returned [{ee}]"
                )            
        except Exception as e:
            cls._logger.warning(
                f"SnowConnect.is_session_active(): Warning {e}!"
            )
        return rb


    def is_current_active(self, session: Session = None) -> bool:
        """
        Check whether current Snowflake connection (shared session) is
        still active or not. 
        
        Returns:
            bool: True if the connection is active, otherwise False
        """
        if session is None:
            session = self._current_connection
        if session is None:
            self.logger.warn(
                f"SnowConnect.is_current_active(): Session is None!"
            )
            return False
        
        rb = SnowConnect.is_session_active(session)
        return rb
    

    def create_user_session(self, ctx: Dict = {}) -> Session:
        """
        Create snowflake user specific session, typically through
        OAuth Auth code/device code flow.

        Arg:
            ctx (Dict): OAuth context dictionary including access
                token, etc.
        Returns:
            Session: Snowflake connection's session
        """
        usr: str = ctx.get(ConfigKey.USER.value)
        session: Session = None
        if self.auth_type == self.T_AUTH_EXT_BROWSER:
            if usr is not None and usr:
                self.connect_params[ConfigKey.USER.value] = usr
            session = self._do_externalbrowser_auth(self.connect_params)
        elif self.auth_type == self.T_AUTH_OAUTH:
            session = self._do_oauth(self.connect_params, ctx)
        return session
    

    def get_service_session(self) -> Session:
        """
        Get existing snowflake data service connection.

        Returns:
            Session: Snowflake connection's session
        """
        if (self.connect_type == SnowConnect.T_SNOWFLAKE_CONN) and\
                ((self.auth_type == self.T_AUTH_KEYPAIR) or
                 (self.auth_type == self.T_AUTH_SNOWFLAKE)):
            return self.get_connection(self.connect_key)
        if (self.connect_type == SnowConnect.T_SNOWFLAKE_CONN) and\
                (self.auth_type == self.T_AUTH_OAUTH) and \
                (self.oauth_flow_type == AppConfig.T_OAUTH_CREDS):
            return self.get_connection(self.connect_key)
        else:
            self.logger.error(
                    "SnowConnect.create_service_session(): Error -"\
                    f"OAuth Client Credentials not yet implemented!")
            return None


    def create_service_session(self) -> Session:
        """
        Create snowflake data service connection.
        
        Returns:
            Session: Snowflake connection's session
        """
        if (self.connect_type == SnowConnect.T_SNOWFLAKE_CONN) and\
                (self.auth_type == self.T_AUTH_KEYPAIR):
            return self.create_connection(self.connect_params)
        
        if (self.connect_type == SnowConnect.T_SNOWFLAKE_CONN) and\
                (self.auth_type == self.T_AUTH_OAUTH) and\
                (self.oauth_flow_type == AppConfig.T_OAUTH_CREDS):
            return self.create_connection(self.connect_params)
        elif (self.connect_type == SnowConnect.T_SNOWFLAKE_CONN) and\
                (self.auth_type == self.T_AUTH_OAUTH):
            raise ValueError(
                    "SnowConnect.create_service_session(): Error -"\
                    f"OAuth Client Credentials not yet implemented!")


    def create_session(self, ctx: Optional[Dict] = {}) -> Session:
        """
        Create snowflake user specific session, or get the existing
        service data connection. Note, use create_user_session() for
        auth code/device code based oauth based user session, and 
        create_service_session() for key based or client credentials
        flow of oauth service session.

        Arg:
            ctx (Dict): OAuth context dictionary including access
                token, etc.
        Returns:
            Session: Snowflake connection's session
        """
        conn = self.get_connection()
        if (conn is not None) and isinstance(conn, Session):
            return conn
        else:
            conn = self.create_connection(self.connect_params)

        session = conn
        if session is None:
            if self.auth_type == self.T_AUTH_EXT_BROWSER:
                session = self._do_externalbrowser_auth(self.connect_params)
            elif self.auth_type == self.T_AUTH_OAUTH:
                if not ctx:
                    cc: ClientCredsConnect = self.oauth_connect
                    d_rs: Dict = cc.grant_request(self.oauth_connect_config)
                    ctx = cc.decode_token(d_rs, ["access_token"])
                
                self.logger.debug(
                        "SnowConnect.create_session()): OAuth "\
                        f"context => {ctx}."
                    )
                session = self._do_oauth(self.connect_params, ctx)
        return session
    

    def close_session(self):
        """
        Close current snowflake connection and session.
        """
        self.close_connection()


    def _create_session(self, auth_type: str, conn_params) -> Session:
        session = None
        try:
            self.logger.debug(
                "SnowConnect._create_session(): Session creation "\
                f"- Auth_type [{auth_type}]; Connect_params => {conn_params}."
            )
            session = Session.builder.configs(conn_params).create() 
            c = session.sql("select current_warehouse(), current_role()")\
                .collect()
            self.logger.debug(
                "SnowConnect._create_session(): Session creation "\
                f"[OK]; Warehouse, Role: [{c}]"
            )
        except Exception as ex:
            self.logger.exception(
                "SnowConnect._create_session(): Cannot connect to "\
                f"Snowflake! Trying auth_type [{auth_type}]; Error [{ex}]!"
            )
        return session


    def _do_snowflake_auth(self, params: Dict):
        pass_env: str = params["password_env"]
        password = os.environ[pass_env]
        conn_params = {
            "account": params["account"],
            "user": params.get("user", ""),
            "authenticator": self.T_AUTH_SNOWFLAKE,
            "password": password,
            "role": params["role"], 
            "warehouse": params["warehouse"],
            "database": params["database"],
            "schema": params["schema"]
        }
        return self._create_session(
            SnowConnect.T_AUTH_SNOWFLAKE, conn_params
        )


    def _do_keypair_auth(self, params: Dict):
        session, pkey, pkb, conn_params = None, None, bytes(), None
        kphrase_env = params["private_key_phrase_env"]
        kphrase = os.environ[kphrase_env]
        kpath_env = params["private_key_path_env"]
        kpath = os.environ[kpath_env]
        try:
            with open(kpath, "rb") as key:
                pkey = serialization.load_pem_private_key(
                    key.read(),
                    password = kphrase.encode(),
                    backend = default_backend()
                )
        except Exception as e:
            self.logger.exception(
                f"SnowConnect._do_keypair_auth(): Cannot read snowflake "\
                f"user private key file: {e}!"
            )
        if pkey:
            try:
                pkb = pkey.private_bytes(
                    encoding = serialization.Encoding.DER,
                    format = serialization.PrivateFormat.PKCS8,
                    encryption_algorithm = serialization.NoEncryption()
                )
            except Exception as e:
                self.logger.exception(
                    f"SnowConnect._do_keypair_auth(): Cannot read snowflake "\
                    f"user private key as bytes: {e}!"
                )
        else: 
            raise ValueError(
                "SnowConnect._do_keypair_auth(): Cannot read snowflake user "\
                "private key file; Check exception and key file path "\
                f"[{kpath}]!"
            )
        if pkb:
            conn_params = {
                "account": params["account"],
                "user": params.get("user", ""),
                "private_key": pkb,
                "role": params["role"], 
                "warehouse": params.get("warehouse", ""),
                "database": params.get("database", ""),
                "schema": params.get("schema", ""),
                "client_session_keep_alive": True,
                "max_connection_pool" : 20
            }
            session = self._create_session(
                SnowConnect.T_AUTH_KEYPAIR, conn_params
            )
        else: 
            raise ValueError(
                "SnowConnect._do_keypair_auth(): Cannot read snowflake user "\
                "private key; Keypair Auth Failed!"
            )
        return session


    def _do_externalbrowser_auth(self, params: Dict):
        conn_params = {
            "account": params["account"],
            "user": params.get("user", ""),
            "authenticator": self.T_AUTH_EXT_BROWSER,
            "role": params["role"], 
            "warehouse": params.get("warehouse", ""),
            "database": params.get("database", ""),
            "schema": params.get("schema", "")
        }
        return self._create_session(
            SnowConnect.T_AUTH_EXT_BROWSER, conn_params
        )


    def _do_oauth(self, params: Dict, ctx: Dict):
        token = ctx.get("access_token")
        dtok = ctx.get("decoded_access_token")

        if dtok is None:
            dtok = params.get("decoded_access_token")
            token = params.get("access_token")
        
        if dtok is not None and dtok:
            upn:str = dict(dtok).get("upn")
            if upn is None:
                upn = params.get(ConfigKey.USER.value, "")
            
            if (upn is None) or (not upn):
                upn = dict(dtok).get("sub")

            upn = upn.upper()
            self.logger.debug(
                f"SnowConnect._do_oauth(): Access_token =>\n"\
                f"{token}\nDecoded_token =>\n{dtok}\n"
            )

            conn_params = {
                "account": params["account"],
                "user": upn,
                "authenticator": AppConnect.T_AUTH_OAUTH,
                "token" : token,
                "role": params["role"], 
                "warehouse": params.get("warehouse", ""),
                "database": params.get("database", ""),
                "schema": params.get("schema", "")
            }
            return self._create_session(
                self.T_AUTH_OAUTH, conn_params
            )
    

    def is_service_connect(self) -> bool:
        """
        Return whether this application connect is service type
        """
        auth_t = self.auth_type
        rs = False
        if auth_t == AppConnect.T_AUTH_EXT_BROWSER:
            rs = False
        elif auth_t == AppConnect.T_AUTH_KEYPAIR:
            rs = True
        elif (auth_t == AppConfig.T_OAUTH) and \
                (self.oauth_flow_type == AppConfig.T_OAUTH_CREDS):
            rs = True
        elif (auth_t == AppConfig.T_OAUTH):
            rs = False
            
        self.logger.debug(
                f"SnowConnect.is_service_connect(): AuthType "\
                f"[{self.auth_type}]; Is_service_connect [{rs}]."
            )
        return rs
    

    @staticmethod
    def ddl(session: Session, sql: str) -> List[Row]:
        """
        Excute Data Definition Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {
            'create', 'drop', 'alter', 'truncate', 'delete', 'replace',
            'rename', 'copy', 'clone', 'show', 'desc', 'undrop',
            'set', 'unset'
        }
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.ddl(): Input sql doesn't seem to be DDL [{sql}]"
            )             
        return session.sql(sql).collect()


    @staticmethod
    def dml(session: Session, sql: str) -> List[Row]:
        """
        Excute Data Manupilation Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {'insert', 'update', 'delete', 'call', 'explain'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.dml(): Input sql doesn't seem to be DML [{sql}]"
            )             
        return session.sql(sql).collect()


    @staticmethod    
    def tcl(session: Session, sql: str) -> List[Row]:
        """
        Excute Tranction Control Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {'begin', 'commit', 'rollback', 'savepoint', 'release'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.tcl(): Input sql doesn't seem to be TCL [{sql}]"
            )             
        return session.sql(sql).collect()


    @staticmethod    
    def dcl(session: Session, sql: str) -> List[Row]:
        """
        Excute Data Control Language statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            List[Row]: list of Snowflake Snowpark Rows
        """
        cmds =  {
            'grant', 'revoke', 'create', 'drop', 'alter', 'use', 'show',
            'desc', 'set', 'unset'
        }
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.dcl(): Input sql doesn't seem to be DCL [{sql}]"
            )             
        return session.sql(sql).collect()
    

    @staticmethod
    def dql(session: Session, sql: str) -> Iterator[DF]:
        """
        Excute Data Query Language (Select) statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            Iterator[DF]: Iterator of Pandas DataFrame list
        """
        cmds =  {'select'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.dql(): Input sql doesn't seem to be DQL [{sql}]"
            )             
        return session.sql(sql).to_pandas_batches()
    

    @staticmethod
    def sdf(session: Session, sql: str) -> SDF:
        """
        Excute Data Query Language (Select) statement in Snowflake
        Args:
            sql (str): sql statement string

        Returns:
            Dataframe: Snowflake Dataframe as result of the query
        """
        cmds =  {'select'}
        matches = [c for c in cmds if sql.upper().startswith(c.upper())]
        if not matches:
            raise ValueError(
                f"SnowConnect.sdf(): Input sql doesn't seem to be DQL [{sql}]"
            )
        return session.sql(sql).to_df()


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains SnowSetup class specific for setting-up, 
initialization, configuration, and tearing-down of Snowflake environment
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


from typing import Optional, Dict, Optional
from snowflake.snowpark import DataFrame as SDF

from snowflake_ai.common import AppConfig
from snowflake_ai.connect import SnowConnect
from snowflake_ai.snowpandas import DataSetup



class SnowSetup(DataSetup):
    """
    This class extends DataSetup for setting-up, initialization,
    configuration, and tearing-down Snowflake specific Environment.

    To use this class, instantiate SnowSetup with SnowConnect as follows:

        >>> from snowflake_ai.common import SnowConnect
        >>> from snowflake_ai.snowpandas import SnowSetup
        ... 
        >>> conn: SnowConnect = SnowConnect()
        >>> setup: SnowSetup = SnowSetup(conn)
        >>> setup.create_stage()
    """
    def __init__(
            self, datasetup_key: str, 
            connect: SnowConnect,
            data: Optional[Dict[str, SDF]] = {},
            app_config: AppConfig = None
    ) -> None:
        super().__init__(
            datasetup_key, 
            connect, 
            {key: sdf.to_pandas() for key, sdf in data.items()},
            app_config
        )
        self._snow_dfs = data


    def get_connect(self) -> SnowConnect:
        """
        Get snowflake connection for the setup.

        Returns:
            SnowConnect: Snowflake data connection
        """
        if isinstance(self._connect, SnowConnect):
            return self._connect
        else:
            raise TypeError(
                "SnowSetup should be initialized with SnowConnect"
            )


    def create_stage(
        self, 
        stage_name: Optional[str] = None
    ):
        """
        Create a snowflake stage in the database referenced by
        SnowConnect configuration.

        Args:
            stage_name (str): snowflake stage name. if none or empty,
                use SnowConnect config parameter     

        Returns:
            None: succesful creation of snowflake stage
        """
        if (stage_name is not None) and (stage_name):
            ddl = f"CREATE STAGE IF NOT EXISTS {stage_name}"
            self.get_connect().ddl(ddl)
        



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains Device code flow OAuthConnect class.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import time
import os
import requests
import sys
import re
from typing import Optional, Dict, Tuple, List
import logging
import streamlit as st
from urllib.parse import urlencode

from snowflake_ai.common import AppConfig, OAuthConnect



class DeviceCodeConnect(OAuthConnect):
    """
    This class represents anapplication connect for 
    OAuth Device code grant flow.

    To use this class, create an instance as following :

        >>> from snowflake_ai.connect import DeviceCodeConnect
        ...
        >>> connect = DeviceCodeConnect()
        >>> connect.authorize_request()
        >>> connect.grant_request()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))


    K_CODE_VERIFIER = "code_verifier"
    K_CODE_CHALLENGE = "code_challenge"
    K_DEVICE_CODE = "device_code"
    T_CODE = "code"
    T_ACCESS_TOKEN = "access_token"
    T_REFRESH_TOKEN = "refresh_token"


    _initialized = False
    _init_connect_params = {}


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = DeviceCodeConnect._logger


    def authorize_request(self, add_params: Dict = {}) -> Dict:
        """
        Make an authorization request. This method is overridden 

        Args:
            add_params (Dict): dictionary contains code challenge for PKCE 
                (Proof Key for Code Exchange) 

        Returns:
            str: url string.
        """ 
        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.auth_request_url
        )

        params = {
            "client_id": self.client_id,
            "scope": self.scope,
        }

        try:
            response = requests.post(url, data=params)
            response.raise_for_status()
            return response.json()
        except requests.HTTPError as http_err:
            self.logger.error(
                f"DeviceCodeConnect.authorize_request(): HTTP error"\
                f" occurred: {http_err}!"
            )
            return {}
        except Exception as err:
            self.logger.error(
                f"DeviceCodeConnect.authorize_request(): Other error"\
                f" occurred: {err}!"
            )
            return {}
    

    def process_authorize_response(self, json_ctx: Dict) -> Dict:
        """
        Construct an authorization request url. This method should be 
        overwritten by its child class.

        Args:
            json_ctx (Dict): json context in form of dictionary

        Returns:
            Dict: input json context
        """
        device_code = json_ctx['device_code']
        user_code = json_ctx['user_code']
        verification_uri = json_ctx['verification_uri']
        expires_in = json_ctx['expires_in']
        interval = json_ctx.get('interval', 10)

        print(f"Please go to {verification_uri} and enter this user"\
                f"code:\n{user_code}")
        self.logger.debug("DeviceCodeConnect.process_authorize_response(): "\
                f"Device_code [{device_code}]; Expires [{expires_in}]")

        ctx = self.prepare_grant_request(json_ctx)
        url = ctx.get("url", "")
        payload = ctx.get("payload", {})
        headers = ctx.get("headers", {})

        for _ in range(0, expires_in, interval):
            response = requests.post(url, data=payload, headers=headers)
            if response.status_code == 200:
                a_token = dict(response.json()).get('access_token', '')
                json_ctx["access_token"] = a_token
                r_token = dict(response.json()).get('refresh_token', '')
                json_ctx["refresh_token"] = a_token
                break
            elif response.status_code == 400:
                print('Waiting for user to authorize...')
            else:
                print('OAuth Device Code Flow Error Occurred:', 
                        response.json())
                break

            time.sleep(interval)
        
        return json_ctx
    

    def prepare_grant_request(self, add_params = {}) -> Dict:
        """
        Prepare grant token request
        """
        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.grant_token_request_url
        )

        code = add_params.get(self.K_DEVICE_CODE)

        if not code:
            raise ValueError(
                "DeviceCodeConnect.grant_request(): Missing required params "\
                "- code!"
        )
        payload = {
            "client_id": self.client_id,
            "scope": self.scope,
            "grant_type": self.grant_type,
            "device_code": code,
        }

        headers = {"Content-Type": self.content_type}
        self.logger.debug(
            f"DeviceCodeConnect.prepare_grant_request(): Payload => {payload}"
        )

        add_params["url"] = url
        add_params["headers"] = headers
        add_params["payload"] = payload

        return add_params


    def grant_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get access token.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        return {}
    

    def prepare_token_refresh(self, ctx: Dict) -> Dict: 
        add_params = {
            DeviceCodeConnect.T_REFRESH_TOKEN: ctx["refresh_token"],
            DeviceCodeConnect.K_CLIENT_SECRET: self.client_secret_env
        }
        return add_params


    def refresh_token_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get refresh token.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        url = re.sub(
            "\{.*?\}",f"{self.tenant_id}", self.grant_token_request_url
        )

        refresh_tok = add_params.get(self.T_REFRESH_TOKEN)
        sec_env = add_params.get(self.K_CLIENT_SECRET)
        secret = os.environ[str(sec_env)]

        payload = {
            "client_id": self.client_id,
            "scope": self.scope,
            "grant_type": self.T_REFRESH_TOKEN,
            "refresh_token": refresh_tok,
            "client_secret": secret
        }

        headers = {"Content-Type": self.content_type}
        self.logger.debug(
            f"DeviceCodeConnect.refresh_token_request(): Payload => {payload}"
        )

        try:
            response = requests.post(url, data=payload, headers=headers)
            response.raise_for_status()
            return response.json()
        except requests.HTTPError as http_err:
            self.logger.error(
                f"DeviceCodeConnect.refresh_token_request(): "\
                f"HTTP error occurred: {http_err}!"
            )
            return {}
        except Exception as err:
            self.logger.error(
                f"DeviceCodeConnect.refresh_token_request(): "\
                f"Other error occurred: {err}!"
            )
            return {}


    def setup_connect(self, params: Optional[Dict] = None):
        """
        Setup OAuthConnect with proper parameters initialized; it
        should be extended by its child class.

        Args:
            params (Dict): input parameters for connection setup.

        Returns:
            OAuthConnect: self with proper parameters initialized
        """
        super().setup_connect(params)
        return self



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


import numpy as np



class Metrics:


    @staticmethod
    def mean_absolute_percentage_error(y_true, y_pred): 
        y_true, y_pred = np.array(y_true), np.array(y_pred)
        return np.mean(np.abs((y_true - y_pred) / y_true)) * 100


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains AppTab class representing an application page tab.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


from typing import Optional, List, Dict, Callable
from functools import wraps

from snowflake.snowpark.session import Session
import streamlit as st
from streamlit import delta_generator as dg



class PageTab:
    """
    Represents a pagetab in a multi-page Streamlit application.

    Attributes:
        tab_id (str): A unique identifier for the page.
        func (Callable): The function that renders the page content.
        tab (DelaGenerator): streamlit tab container.
        title (str): The title of the page displayed in the navigation.

    Example:
        >>> def page_home():
        ...     st.title("Home")
        ...     st.write("Welcome to the Home page!")
        ...
        >>> home_page = AppPage("home", "Home", page_home)
    """

    def __init__(
        self, 
        tab_id: str,
        func: Callable,
        tab: dg.DeltaGenerator = None,
        title: Optional[str] = None
    ):
        self.tab_id = tab_id
        self.func = func
        self._tab = tab
        if title is None:
            self.title = tab_id
        else:
            self.title = title


    def get_tab(self) -> dg.DeltaGenerator:
        """
        Get the underlying streamlit tab object

        Return:
            DeltaGenerator: Underlying tab object from streamlit for
                this AppTab object
        """
        return self._tab


    def set_tab(self, tab:dg.DeltaGenerator):
        """
        Set the underlying streamlit tab object.

        Args:
            tab (DeltaGenerator): Underlying tab object to be set
        """
        self._tab = tab
        return self



class TabRegistry:
    """
    A registry for managing tabs in a multi-page application.

    Attributes:
        tabs (List[PageTab]): A list of PageTab instances.

    Example:
        >>> hp = HomePage()
        >>> register = hp.tab_regitry.register
        >>> @register("home", "Home")
        ... def home_tab():
        ...     st.title("Home")
        ...     st.write("Welcome to the Home page!")
    """

    def __init__(self):
        self.tab_list: List[PageTab] = []
        self.tab_funcs: List[Callable] = []
        self.tabs: Dict[str, PageTab] = {}


    def register(
        self,
        tab_id: str,
        title: Optional[str] = None
    ) -> Callable:
        """
        Register a tab in the application page.

        Args:
            tab_id (str): A unique identifier for the tab.
            title (Optional[str]): The title of the tab displayed in the
                navigation. Defaults to None.

        Returns:
            Callable: The wrapped page tab function.

        Example:
            >>> registry = TabRegistry()
            >>> @registry.register("home", "Home")
            ... def home_tab():
            ...     st.title("Home")
            ...     st.write("Welcome to the Home page!")
        """

        def decorator(func: Callable) -> Callable:
            @wraps(func)
            def wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            pt = PageTab(tab_id, wrapper, None, title)
            self.tab_list.append(pt)
            self.tab_funcs.append(func)
            return wrapper

        return decorator


    def create_page_tabs(self) -> Dict:
        """
        Create a dictionary of PageTab objects with their underlying
        streamlit container objects set.

        Return:
            dict: PageTab object ditionary
        """
        ids = [t.tab_id for t in self.tab_list]
        tbs = st.tabs(ids)

        td = dict(zip(ids, tbs))
        rd = {}

        for pt in self.tab_list:
            if pt.get_tab() is None and pt.tab_id in td:
                pt.set_tab(td[pt.tab_id])
                rd[pt.tab_id] = pt

        self.tabs = rd
        return rd


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains SnowEncoder and its child classes to encode column.

"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


from typing import (
    Dict, Union, Optional, List
)

from pandas import DataFrame as DF
import json
from sklearn.preprocessing import LabelEncoder
from snowflake.snowpark.functions import udf
from snowflake.snowpark.types import IntegerType

from snowflake.snowpark import DataFrame as SDF
from snowflake.snowpark.functions import col, object_construct, CaseExpr
from snowflake.snowpark.functions import lit, array_agg, to_varchar, when
from snowflake.snowpark.functions import sql_expr, iff

from snowflake_ai.snowpandas import SnowPrep



class SnowEncoder:
    T_PROC_DROP = -1
    T_PROC_SKIP = 0
    T_PROC_GEN = 1

    def __init__(
            self,
            input_cols: Optional[Union[List[str], str]] = None,
            output_cols: Optional[Union[Dict, str]] = None,
            categories: Optional[Dict] = None,
            process_flag: int = 1
    ):
        """
        Encode categorical features using one-hot encoding
        
        Parameters:
            input_cols: name of column or list of columns to encode
            output_cols: name of output column or list of output 
                columns,need to be in the same order as the categories
            categories: {} or specified as a dict: 
                {"COL1": [cat1, cat2, ...], "COL2": [cat1, cat2, ...], ...}
            proc_flags: 
                -1 - drop columns
                 0 - skip columns keeping as it is
                 1 - default to generate matching unknown columns
        """
        self.input_cols = input_cols
        self.output_cols = output_cols
        self.fitted_values = {}
        self.categories = categories
        self.process_flag = process_flag


    def fit(self, df: SDF) -> 'SnowEncoder':
        pass


    def transform(self, df: SDF) -> SDF:
        pass


    def check_fitted(encoder):
        if not hasattr(encoder, "fit"):
            raise TypeError(f"{encoder} is not an encoder instance.")

        fitted = [
            v for v in vars(encoder) if v.endswith("_")
        ]
        if not fitted:
            raise TypeError(f"{type(encoder).__name__} not fitted.")


    def check_columns(columns:List[str], df:SDF):
        df_columns = set(df.columns)
        target_cols = set([col.upper() for col in columns])

        missing_cols = target_cols - df_columns
        if len(missing_cols):
            raise ValueError(
                f"Cannot find columns {missing_cols} in the input dataframe."
            )


    def encode_column_expressions(encoder):
        """
        Builds a list of Column expressions that map categories in columns
        to integer indices. 

        Parameters:
            encoder: An object containing 'input_cols' (columns to be encoded), 
            'fitted_values_' (unique categories per column), 
            and optionally 'handle_unknown' (a strategy to handle unknown 
            categories) and 'unknown_value' (value to replace unknown 
            categories).

        Returns:
            List[Column]: A list of PySpark Column expressions defining the 
                encoding for each column.
        """
        col_exprs = []
        input_cols = encoder.input_cols
        fitted_cats = encoder.fitted_values

        for column in input_cols:
            case_expression = None
            for index, category in enumerate(fitted_cats[column]):
                if isinstance(case_expression, CaseExpr):
                    case_expression = case_expression.when(
                        col(column) == lit(category), lit(index)
                    )
                else:
                    case_expression = when(
                        col(column) == lit(category), lit(index)
                    )

            # set to 'use_encoded_value' if 'handle_unknown' attribute exists
            if hasattr(encoder, 'handle_unknown') and \
                    encoder.handle_unknown == "use_encoded_value":
                case_expression = case_expression.otherwise(
                    lit(encoder.unknown_value)
                )

            col_exprs.append(case_expression)

        return col_exprs


    def decode_column_to_sql_expressions(encoder):
        """
        Generates a list of SQL expressions to map encoded indices back
        to original categories.

        Parameters:
        encoder: An object containing 'output_cols' (columns after encoding),
             'fitted_values_' (unique categories per column), 
            and 'input_cols' (original columns before encoding).

        Returns:
        List[Column]: A list of PySpark SQL expressions defining the decoding 
            for each column.
        """
        input_columns = encoder.input_cols
        output_columns = encoder.output_cols
        fitted_categories = encoder.fitted_values

        # Create tuples of corresponding input and output columns
        column_pairs = list(zip(input_columns, output_columns))
        sql_expressions = []
        
        for input_col, output_col in column_pairs:
            # Convert fitted categories to JSON and create SQL expression
            json_categories = json.dumps(fitted_categories[input_col])
            sql_expression = sql_expr(
                f"AS_CHAR(PARSE_JSON('{json_categories}')[{output_col}])"
            )

            sql_expressions.append(sql_expression)

        return sql_expressions


    def encode_column_names(self):
        """
        Generate encoded column names for categorical variables.
        """

        cat_cols = {}

        if self.output_cols:
            if len(self.output_cols) != len(self.input_cols):
                raise ValueError(
                    f"Number of output columns ({len(self.output_cols)}) "\
                    f"should match number of input columns "\
                    f"({len(self.input_cols)})."
                )

            total_categories = sum(len(self.fitted_values[col]) \
                    for col in self.input_cols)
            total_output_cols = sum(len(self.output_cols[col]) \
                    for col in self.output_cols)

            if total_categories != total_output_cols:
                raise ValueError(
                    f"Total categories ({total_categories}) should match "\
                    f"total output columns ({total_output_cols}).")

            cat_cols = self.output_cols

        else:
            # Generate encoded column names
            for col in self.input_cols:
                cat_cols[col] = [
                    f'{col}_{val}' for val in self.fitted_values[col]
                ]

        total_cols = sum(len(cat_cols[col]) for col in cat_cols)
        if total_cols > 3650:
            raise ValueError("Too many columns; use other encoder instead.")

        return cat_cols


    def fit_transform(self, df: SDF) -> SDF:
        """
        Fit with this encoder and transform the dataframe.

        Parameters:
            df: DataFrame to encode

        Return:
            Encoded DataFrame
        """
        return self.fit(df).transform(df)    


class OneHotEncoder(SnowEncoder):

    def fit(self, df: SDF) -> 'OneHotEncoder':
        """
        Fit the OneHotEncoder using df.

        Parameters:
            df: Snowflake DataFrame used for getting the categories for
                each input column
        
        Return: 
            Fitted encoder
        """
        encode_cols = SnowPrep.check_input_columns(df, self.input_cols)
        self.input_cols = encode_cols

        self.fitted_values = SnowPrep.get_columns_unique_categories(
            df, self.categories, encode_cols)

        return self
    

    def transform(self, df: SDF) -> SDF:
        """
        Transform dataframe df using one-hot encoding creating one new 
        column for each category found with fit.

        If process_flag is -1 then the input columns are dropped

        Parameters:
            df: Snowpark DataFrame to transform
        
        Return: 
            Encoded Snowpark DataFrame
        """
        self.check_fitted()
        out_cols = self.encode_column_names()
        self.output_cols = out_cols

        for c in self.input_cols:
            uniq_vals = self.fitted_values[c]
            col_names = out_cols[c]
            df = df.with_columns(
                col_names, 
                [iff(col(c) == val, lit(1), lit(0)) for val in uniq_vals]
            )
            if self.process_flag == self.T_PROC_GEN:
                df = df.with_column(
                    f"{c}__unknown", 
                    iff(~ col(c).in_(uniq_vals), lit(1), lit(0))
                )

            if self.process_flag == self.T_PROC_DROP:
                df = df.drop(c)

        return df



class BinaryEncoder(SnowEncoder):
    """
    A binary encoder that encodes categorical features using a 
    binary scheme.
    """

    def fit(self, df: SDF) -> 'BinaryEncoder':
        """
        Fit the BinaryEncoder using df.

        Parameters:
            df: DataFrame used for getting the categories for each input
                column
        Return: 
            Fitted encoder
        """
        encode_cols = SnowPrep.check_input_columns(df, self.input_cols)
        self.input_cols = encode_cols

        self.fitted_values = {}
        for col in self.input_cols:
            categories = SnowPrep.get_columns_unique_categories(df, None, [col])
            label_encoder = LabelEncoder()
            label_encoder.fit(categories)
            self.fitted_values[col] = label_encoder

        return self


    def transform(self, df: SDF) -> SDF:
        """
        Transform dataframe df using binary encoding, creating new columns for 
        each bit of the binary representation of the category.

        If process_flag is -1 then the input columns are dropped

        Parameters:
            df: DataFrame to transform

        Return: 
            Encoded DataFrame
        """
        self.check_fitted()

        for c in self.input_cols:
            label_encoder:LabelEncoder = self.fitted_values[c]
            
            # Define UDF to apply the label encoder and transform to binary
            encode_udf = udf(
                lambda category: label_encoder.transform([category])[0], 
                IntegerType()
            )
            
            df = df.withColumn(c, encode_udf(df[c]))

            if self.process_flag == self.T_PROC_DROP:
                df = df.drop(c)

        return df


class OrdinalEncoder(SnowEncoder):
    """
    An ordinal encoder that encodes categorical features using an 
    ordinal scheme.
    """

    def fit(self, df: SDF) -> 'OrdinalEncoder':
        """
        Fit the OrdinalEncoder using df.

        Parameters:
            df: DataFrame used for getting the categories for each input
                column
        Return: 
            Fitted encoder
        """
        encode_cols = SnowPrep.check_input_columns(df, self.input_cols)
        self.input_cols = encode_cols

        self.fitted_values = {}
        for col in self.input_cols:
            categories = SnowPrep.get_columns_unique_categories(df, None, [col])
            label_encoder = LabelEncoder()
            label_encoder.fit(categories)
            self.fitted_values[col] = label_encoder

        return self


    def transform(self, df: SDF) -> SDF:
        """
        Transform dataframe df using ordinal encoding, creating new columns 
        for each category in ordinal form.

        If process_flag is -1 then the input columns are dropped

        Parameters:
            df: DataFrame to transform

        Return: 
            Encoded DataFrame
        """
        self.check_fitted()

        for c in self.input_cols:
            label_encoder:LabelEncoder = self.fitted_values[c]
            
            # Define UDF to apply the label encoder
            encode_udf = udf(
                lambda category: label_encoder.transform([category])[0], 
                IntegerType()
            )
            
            df = df.withColumn(c, encode_udf(df[c]))

            if self.process_flag == self.T_PROC_DROP:
                df = df.drop(c)

        return df



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains DataConnect class representing a generic data
connection configuration.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import sys
from typing import Optional, Dict, Union
import logging

from snowflake.snowpark import Session
from snowflake_ai.common import ConfigType, ConfigKey
from snowflake_ai.common import AppConfig, AppConnect



class DataConnect(AppConnect):
    """
    This class represents a generic data connection.

    Typically, you don't use this class to create data connection
    directly, rather create an instance from AppConfig or 
    extend it in a child class for creation of a specific data
    connection instance, e.g., SnowConnect as following :

        >>> from snowflake_ai.connect import SnowConnect
        ...
        ... # using default snowflake config directly
        >>> connect: SnowConnect = SnowConnect()
    """

    K_DATA_CONN = ConfigType.DataConnects.value
    K_INIT_LIST = ConfigKey.INIT_LIST.value

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _initialized = False

    # {<connect_key>: file-connect-string|snowflake-connection-obj}
    _data_connections = {}


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = DataConnect._logger
        self._current_connection = None

        # load default data service connection
        if (connect_key is None or not connect_key or \
                self._current_connection is None) and self.data_connections:
            k = DataConnect.search_default_key(self.data_connections)
            self.connect_group, self.connect_key = \
                AppConfig.split_group_key(k)
            self.connect_key, self.connect_params = \
                AppConnect.load_connect_config(k, self._configs)
            self.app_connects[self.connect_key] = self
            self.set_current_connection(self.data_connections[k])
            connect_key = self.connect_key

        # load user based connection
        if not connect_key:
            self.logger.error(f"DataConnect.init(): Error - missing key"\
                    f" [{connect_key}]; Connect_key[{self.connect_key}]")
        elif not self.connect_key:
            self.connect_key = connect_key
            self.connect_group, self.connect_name = \
                AppConfig.split_group_key(connect_key)
       
        self.logger.debug(f"DataConnect.init(): Connect_group"\
                f"[{self.connect_group}]; Connect_key[{self.connect_key}]"\
                f"; Connect_name[{self.connect_name}]")

        # setup oauth reference
        if self.connect_params:
            self.connect_type = self.connect_params.get(
                    ConfigKey.TYPE.value, '')
            self.connect_name = self.connect_params.get(
                    ConfigKey.NAME.value, '')
            self.oauth_connect_ref = self.connect_params.get(
                    ConfigKey.CONN_OAUTH.value)
            if self.oauth_connect_ref is not None and \
                    self.oauth_connect_ref:
                self.auth_type = AppConfig.T_OAUTH
            else:
                self.oauth_connect_ref = ""
            
            if self.oauth_connect_ref:
                _, self.oauth_connect_config = \
                        AppConfig.get_group_item_config(
                                self.oauth_connect_ref, 
                                ConfigType.AppConnects.value,
                                self._configs)
            else:
                self.oauth_connect_config = {}

            self.logger.debug(f"DataConnect.init(): Snowflake OAuth "\
                    f" Configuration => {self.oauth_connect_config}.")

            # match oauth connect connect_type
            if self.oauth_connect_config:
                self.oauth_flow_type = self.oauth_connect_config.get(
                        ConfigKey.TYPE.value, "")
            


    @property
    def data_connections(self) -> Dict:
        """
        Get a dictionary of all data connections.

        Returns:
            dict: a dictionary of all data connections.
        """   
        return DataConnect._data_connections


    def get_current_connection(self) -> Union[str, Session]:
        """
        Get current data connection.

        Returns:
            str, Session: specific data connection object, e.g. file
                path or Snowflake Session.
        """
        if not self.data_connections:
            self.logger.warning(
                "DataConnect.current_connection(): DataConnect."\
                "_data_connections dictionary is empty!"
            )
        elif (self._current_connection is None):
            k = DataConnect.search_default_key(self.data_connections)
            self.connect_group, self.connect_key = \
                AppConfig.split_group_key(k)
            self.connect_key, self.connect_params = \
                AppConnect.load_connect_config(k, self._configs)
            self.app_connects[self.connect_key] = self
            self.set_current_connection(self.data_connections[k])

        if ((not self.is_current_active()) or \
                (self.data_connections.get(self.connect_key) is None)) \
                and (self.is_service_connect()):
            self.data_connections[self.connect_key] = \
                self.create_connection(self.connect_params)
            self.set_current_connection(
                self.data_connections[self.connect_key]
            )
        return self._current_connection
    

    def set_current_connection(self, conn: Union[str, Session]):
        """
        Set current data connection.

        Args:
            conn (str | Session): input connection object
        Returns:
            str, Session: specific data connection object, e.g. file
                path or Snowflake Session.
        """
        if conn is not None:
            self._current_connection = conn
        return self._current_connection


    def is_current_active(self) -> bool:
        """
        Check whether current data connection is active or not. It should
        be overwritten by specific subclass as it is implementation
        dependant.

        Returns:
            bool: True if it is active, otherwise False
        """
        if self._current_connection is None:
            return False
        elif not self._data_connections:
            return False
        elif self.is_oauth_saml_type():
            return False
        elif self.connect_type == AppConnect.T_SNOWFLAKE_CONN:
            return False
        return True
    

    def create_connection(self, params: Dict):
        """
        Create a data connection based on the specific configuration.
        Child class should implement specific connnection logic
        to override this method.

        Args:
            params (Dict): input parameters for connection creation.

        Returns:
            object: data connection object, e.g., connection session for
                snowflake connection.
        """
        conn = f"Connected to {params['host']}:{params['port']}"
        return conn


    def close_connection(self):
        """
        Close current data connection.
        
        Returns:
            int: 0 - successful; otherwise unsuccessful
        """
        rn = 0
        try:
            if self._current_connection is not None and \
                    isinstance(self._current_connection, Session):
                self._current_connection.close()
        except Exception as e:
            self.logger.error(f"DataConnect.close_connection(): Error {e}")
            rn = -1
        return rn


    def get_connection(self, connect_key: Optional[str] = None):
        """
        Get lazy created shared data service connection or change 
        existing data connection to use the input connect_key to initialize
        connection. This should not be used for user specific data
        connection.

        Args:
            connect_key (str): data connect key in form of 
                "data_connects".<connect_name>

        Returns:
            object: data connection object, e.g., session connection for
                snowflake connection; it can return None.
        """
        if connect_key is None:
            if self._current_connection is None:
                if self.data_connections:
                    k = DataConnect.search_default_key(self.data_connections)
                    self.connect_group, self.connect_key = \
                        AppConfig.split_group_key(k)
                    self.connect_key, self.connect_params = \
                        AppConnect.load_connect_config(k, self._configs)
                    self.app_connects[self.connect_key] = self
                    self._current_connection = self.data_connections[k]
                    self.logger.debug(
                        f"DataConnect.get_connection(): Connect_group ["\
                        f"{self.connect_group}]; Connect_key "\
                        f"[{self.connect_key}]; Current_connect_key [{k}]; "\
                        f"Current_connection [{self._current_connection}]."
                    )
                else:
                    self.logger.warning(
                        f"DataConnect.get_connection(): Warning - "\
                        f"[{len(self._data_connections)}] " \
                        "connection initialized!"
                    )
                    return None
            else:
                return self._current_connection
        else:
            qk = AppConfig.get_qualified_key(
                ConfigType.AppConnects.value, connect_key
            )
            if self.data_connections.get(qk) is None:
                gk, k = AppConfig.split_group_key(qk)
                params = self._configs[ConfigType.AppConnects.value]\
                    [DataConnect.K_DATA_CONN][k]
                self.connect_key = qk
                self.connect_params = params
                self.connect_group = gk
                if self.is_service_connect():
                    self.data_connections[qk] = self.create_connection(params)
                    self._current_connection = self.data_connections[qk]

            return self.data_connections.get(qk)


    def init_connects(self) -> int:
        """
        Initialize DataConnect in [app_connects.data_connects] init_list.

        Returns:
            int: 0 - if there is no connect object has been initialized;
                otherwise, return an integer to show the number of 
                AppConnect objects have been initialized.
        """
        if not DataConnect._initialized:
            conn_group_dict = AppConfig.filter_group_key(
                DataConnect.K_DATA_CONN, 
                AppConnect.K_APP_CONN, 
                self._configs
            )
            lst = conn_group_dict.get(DataConnect.K_INIT_LIST)
            if lst is not None and len(lst) > 0 :
                for dconn in lst:
                    dc: dict = self._configs[AppConnect.K_APP_CONN]\
                        [DataConnect.K_DATA_CONN]
                    params = dc.get(dconn)
                    if params is None:
                        raise ValueError(
                            f"DataConnect.init_connets(): Error - [{dconn}]"\
                            " doesn't exist in the configuration!"
                        )
                    elif self.is_service_connect():
                        s = f"{DataConnect.K_DATA_CONN}.{dconn}"
                        if self.data_connections.get(s) is None:
                            self.data_connections[s] = \
                                    self.create_connection(params)
                            self.set_current_connection(
                                    self.data_connections[s]
                                )
                DataConnect._initialized = True
        
        n = len(self.data_connections)
        self.logger.debug(
            f"DataConnect.init_connects(): [{n}] shared (service) "\
            f"data connections have been established."
        )
        return n


    def is_service_connect(self) -> bool:
        """
        Return whether this application connect is service type.
        This should be overridden by its childen class.
        """
        return False


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains AppTab class representing an application page tab.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.2.0"


from typing import Optional, List, Dict, Callable
from functools import wraps

from snowflake.snowpark.session import Session
import streamlit as st
from streamlit import delta_generator as dg

from snowflake_ai.apps import TabRegistry



class AppPage:
    """
    Represents a page in a multi-page Streamlit application.

    Attributes:
        tab_id (str): A unique identifier for the page.
        func (Callable): The function that renders the page content.
        tab (DelaGenerator): streamlit container.
        title (str): The title of the page displayed in the navigation.
        icon (str): The title of the page displayed in the navigation.

    Example:
        >>> def page_home():
        ...     st.title("Home")
        ...     st.write("Welcome to the Home page!")
        ...
        >>> home_page = AppPage("home", "Home", page_home)
    """

    def __init__(
        self,
        page_id: str,
        title: Optional[str] = '',
        icon: Optional[str] = '',
        layout: Optional[str] = "wide",
        show_sidebar: Optional[bool] = True,
        show_tab: Optional[bool] = True,
        session: Session = None,
        func: Optional[Callable] = None,
    ):
        self.page_id = page_id
        if title is None:
            self.title = page_id
        else:
            self.title = title
        self.icon:str = icon
        self.layout:str = layout
        self.show_sidebar = show_sidebar
        self.show_tab = show_tab
        self.tab_registry = TabRegistry()
        self.session = session
        self.func = func


    def link_login(self, url):
        st.markdown(
            f"<a href='{url}' target = '_self'>Login</a>", 
            unsafe_allow_html=True
        )


    def link_logoff(self, url):
        st.markdown(
            f"<a href='{url}' target = '_self'>Logoff</a>", 
            unsafe_allow_html=True
        )


    def add_logo(self):
        st.markdown(
            """
            <style>
                [data-testid="stSidebarNav"] {
                    background-image: url(app/static/ecolab-logo.png);
                    background-repeat: no-repeat;
                    padding-top: 120px;
                    background-position: 8px 140px;
                    background-size: 280px 80px;
                }
                [data-testid="stSidebarNav"]::before {
                    content: "Insight Service";
                    margin-left: 20px;
                    margin-top: 20px;
                    font-size: 30px;
                    position: relative;
                    top: 100px;
                    color: #007AC3;
                }
            </style>
            """,
            unsafe_allow_html=True,
        )
        
        
    def render_sidebar(self):
        """
        Overwritten by its child class
        """
        self.add_logo()
        is_login = getattr(st.session_state, "login", False)
        sts: Dict = st.experimental_get_query_params()
        state = sts.get("state")
        loggedin = False
        if state:
            t = state[0]
            loggedin = True if t == "loggedin" else False
        else:
            loggedin = is_login

        if not loggedin:
            st.experimental_set_query_params(
                page="/", state="loggedoff"
            )
            setattr(st.session_state, "login", False)

        else:
            url = "http://localhost:8501/home"
            with st.sidebar:
                self.link_logoff(url)

            st.experimental_set_query_params(
                page="/", state="loggedin",
            )
            setattr(st.session_state, "login", True)


    def render(self):
        css = '''
            <style>
                .appview-container .main .block-container {{
                    max-width: 100%;
                    padding-top: 0rem;
                    padding-right: 0rem;
                    padding-left: 0rem;
                    padding-bottom: {1}rem; 
                }}

                .stTabs [data-baseweb="tab-list"]
                    button [data-testid="stMarkdownContainer"] 
                    p { font-size: 1.5rem; color: #1D8CCC}

                </style>
            '''
        st.markdown(css, unsafe_allow_html=True)

        if self.show_sidebar:
            self.render_sidebar()

        if self.show_tab:
            self.tab_registry.create_page_tabs()
            for t in self.tab_registry.tab_list:
                with t.get_tab():
                    t.func(self.session)

